"""Step Menu CLI Entrypoints"""

import logging
import sys

import click

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def info(profile: str, token: str, url: str, **kwargs) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO
    """
    # Check if URL is ok
    if not cu.is_full_url(url):
        click.secho(f'INPUT ERROR: Step url is not a URL: {url}', fg='bright_red', bold=True)
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile, token)

    # Request the data
    data = yj_obj.step.info(step_url=url)

    if not data:
        click.secho('No step information', fg='bright_red', bold=True)
        sys.exit(1)
    cu.standard_out(data, **kwargs)



"""Build click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import build
from yojenkins.cli import cli_build, cli_decorators
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@build.command(short_help='\tBuild information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def info(ctx, debug, **kwargs):
    """Build information"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.info(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tBuild status text/label')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number', metavar='INT')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=bool, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def status(ctx, debug, **kwargs):
    """Build status text/label"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.status(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())
    # FIXME: Showing running for job, seems not to have picked the right build ... when url to job is used!


@build.command(short_help='\tAbort build')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def abort(ctx, debug, **kwargs):
    """Abort build"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.abort(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tDelete build')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def delete(ctx, debug, **kwargs):
    """Delete build"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.delete(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build stages')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def stages(ctx, debug, **kwargs):
    """Get build stages"""
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.stages(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build logs')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('--tail', type=float, required=False, help='Last of logs. If < 1 then %, else number of lines')
@click.option('-dd',
              '--download-dir',
              type=click.Path(file_okay=False, dir_okay=True),
              required=False,
              is_flag=False,
              help='Download logs to directory')
@click.option('--follow',
              default=False,
              type=str,
              required=False,
              is_flag=True,
              help='Follow/Stream the logs as they are generated')
@click.pass_context
def logs(ctx, debug, **kwargs):
    """Get build logs

    EXAMPLES:

    \b
    - yojenkins build logs "myFolder/myJob" --latest
    - yojenkins build logs "myFolder/myJob" --latest --tail 10
    - yojenkins build logs "myFolder/myJob" --latest --tail 0.1
    - yojenkins build logs "myFolder/myJob" --number 2 --follow
    - yojenkins build logs "myFolder/myJob" --latest -dd .

    """
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.logs(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tOpen build in web browser')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def browser(ctx, debug, **kwargs):
    """Open build in web browser"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.browser(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tStart monitor UI')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('-s', '--sound', type=bool, required=False, is_flag=True, help='Enable sound effects')
@click.pass_context
def monitor(ctx, debug, **kwargs):
    """Start monitor UI"""
    # TODO: Pass a list of build numbers
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.monitor(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tGet build parameters')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.pass_context
def parameters(ctx, debug, **kwargs):
    """Get build parameters

    Build parameters are the parameters that were used to start the build.
    """
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.parameters(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tRebuild build')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('job', nargs=1, type=str, required=False)
@click.option('-n', '--number', type=int, required=False, help='Build number')
@click.option('-u', '--url', type=str, required=False, help='Flexible build URL (No job info needed)')
@click.option('--latest', type=str, required=False, is_flag=True, help='Latest build (Replaces --number)')
@click.option('--follow-logs',
              type=bool,
              required=False,
              is_flag=True,
              help='Wait for build, follow logs when build starts')
@click.pass_context
def rebuild(ctx, debug, **kwargs):
    """Get build parameters

    Rebuild the specified build exactly as it ran
    """
    set_debug_log_level(debug)
    if kwargs.get("job") or kwargs.get("url"):
        cli_build.rebuild(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@build.command(short_help='\tFind difference between two builds')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('build-url-1', nargs=1, type=str, required=True)
@click.argument('build-url-2', nargs=1, type=str, required=True)
# @click.option('--type',
#               type=click.Choice(['info', 'logs'], case_sensitive=False),
#               default="info",
#               show_default=True,
#               required=False,
#               help='Type of diff comparison')
@click.option('--logs', type=bool, default=False, required=False, is_flag=True, help='Build logs diff')
@click.option('--line-pattern',
              type=str,
              multiple=True,
              required=False,
              help='Keep part of each line that contains this REGEX pattern [Can use multiple times]')
@click.option('--char-ignore',
              default=0,
              type=click.IntRange(0),
              required=False,
              help='Number of characters to ignore at line start')
@click.option('--no-color', type=bool, default=False, required=False, is_flag=True, help='Show output without color')
@click.option('--diff-only',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Show only lines that are different')
@click.option('--diff-guide',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Show where the difference is in line')
# @click.option('--stats-only', type=bool, default=False, required=False, is_flag=True, help='Show diff stats only')
def diff(debug, **kwargs):
    """Get the diff comparison for two builds (info, logs)

    EXAMPLES:

    \b
    - yojenkins build diff "http://jenkins.com/myFolder/myJob/4" "http://jenkins.com/myFolder/myJob/5"
    - yojenkins build diff "<SERVER>/myJob/4/console" "<SERVER>/myJob/5" --logs --diff-guide
    - yojenkins build diff "<SERVER>/myJob/5/" "<SERVER>/yourJob/8" --diff-only
    - yojenkins build diff "<SERVER>/myJob/2/" "<SERVER>/youJob/4/" --logs --char-ignore 40
    - yojenkins build diff "<SERVER>/myJob/2/" "<SERVER>/youJob/4/" --logs --line-patten "(?<= - ).+"
    - yojenkins build diff "<SERVER>/myJob/2/" "<SERVER>/youJob/4/" --logs --line-patten "FAIL" --line-patten "SUCCESS"

    """
    set_debug_log_level(debug)
    cli_build.diff(**translate_kwargs(kwargs))



""""Logger Setup"""

# TODO: Move to utilities

import logging
import sys
from logging.handlers import RotatingFileHandler

import coloredlogs

# Turn off INFO level logging for python package "jons2xml"
logging.getLogger("dicttoxml").setLevel(logging.ERROR)
logging.getLogger("urllib3").setLevel(logging.WARNING)

# Setting up log file handler
file_handler = RotatingFileHandler(filename='yojenkins.log',
                                   mode='a',
                                   maxBytes=5000000,
                                   backupCount=0,
                                   delay=True,
                                   encoding="utf-8")

# Also include any sys.stdout in logs
stdout_handler = logging.StreamHandler(sys.stdout)

# Defining the logger
# NOTE: More items: https://docs.python.org/3/library/logging.html#logrecord-attributes
# NOTE: Excluding %(levelname)-8s
LOG_FORMAT = '[%(asctime)s] [%(relativeCreated)-4d] [%(filename)-24s:%(lineno)4s] %(message)s'

# Basic Configurations
logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt='%H:%M:%S', handlers=[file_handler, stdout_handler])

# Applying color to the output logs
coloredlogs.install(fmt=LOG_FORMAT, datefmt='%H:%M:%S')



"""Stage class definition"""

import logging
import os
import threading
from datetime import datetime, timedelta
from typing import Dict, Tuple, Union

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out, print2
from yojenkins.yo_jenkins.status import StageStatus

# Getting the logger reference
logger = logging.getLogger()


class Stage():
    """Handeling of Jenkins stage functionality"""

    def __init__(self, rest: object, build: object, step: object) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest
        self.build = build
        self.step = step

        self.build_logs_extension = '.log'

        self._stage_log_list_thread_lock = threading.Lock()
        self.stage_log_dict = {}

    def info(self,
             stage_name: str,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: int = None,
             latest: bool = False) -> Dict:
        """Get the stage information for specified stage

        Details: Ways of specifying the build:
            - Build URL only
            - Job URL only
            - Job name with build number or latest flag

        Args:
            stage_name   : Name of the stage
            build_url    : Direct URL of the build
            job_name     : Name of the Job
            job_url      : URL of the Job
            build_number : Build number for the Job
            latest       : Latest build

        Returns:
            Stage information
        """
        if build_url:
            build_url = utility.build_url_complete(build_url)

        # Getting all stages
        build_stage_list, build_stage_name_list = self.build.stage_list(build_url, job_name, job_url, build_number,
                                                                        latest)
        logger.debug(f'Stages found: {build_stage_name_list}')

        # Formate stage name from user input
        logger.debug(f'Formatting stage name "{stage_name}": Lower case, strip spaces')
        stage_name = stage_name.lower().strip()

        # Check if lowercase stage name is in list of stages in this build
        if not stage_name in [x.lower() for x in build_stage_name_list]:
            fail_out(f'Failed to find stage name "{stage_name}" among listed stages')

        # Getting the right stage item
        logger.debug('Getting stage URL from build information ...')
        build_stage_item = next(item for item in build_stage_list if item["name"].lower() == stage_name)

        # Making the request to get stage info
        endpoint = f'{build_stage_item["url"]}'
        return_content = self.rest.request(endpoint, 'get', is_endpoint=True)[0]
        if not return_content:
            fail_out(f'Failed to fetch stage information for "{stage_name}"')

        # Add additional derived information for stage
        return_content['startDatetime'] = datetime.fromtimestamp(return_content["startTimeMillis"] /
                                                                 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
        return_content['durationFormatted'] = str(timedelta(seconds=return_content["durationMillis"] / 1000.0))[:-3]
        return_content['pauseDurationFormatted'] = str(
            timedelta(seconds=return_content["pauseDurationMillis"] / 1000.0))
        return_content['numberOfSteps'] = len(return_content['stageFlowNodes'])

        # Add additional derived information for each step
        if "stageFlowNodes" in return_content:
            # Accounting for no stage step command
            for step_info in return_content['stageFlowNodes']:
                step_info['startDatetime'] = datetime.fromtimestamp(step_info["startTimeMillis"] /
                                                                    1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
                step_info['durationFormatted'] = str(timedelta(seconds=step_info["durationMillis"] / 1000.0))[:-3]
                step_info['pauseDurationFormatted'] = str(timedelta(seconds=step_info["pauseDurationMillis"] / 1000.0))

                # Adding the urls to the item
                step_info['url'] = step_info['_links']['self']['href']
                step_info['url_log'] = step_info['_links']['log']['href']
                step_info['url_console'] = step_info['_links']['console']['href']
                step_info['url_full'] = f'{self.rest.get_server_url()}{step_info["url"]}'

        # TODO: Make utility function for additional derived info

        return return_content

    def status_text(self,
                    stage_name: str,
                    build_url: str = '',
                    job_name: str = '',
                    job_url: str = '',
                    build_number: int = None,
                    latest: bool = False) -> str:
        """Get the status text of the specified stage

        Details: Ways of specifying the build:
            - Build URL only
            - Job URL only
            - Job name with build number or latest flag

        Args:
            stage_name   : Name of the stage
            build_url    : Direct URL of the build
            job_name     : Name of the Job
            job_url      : URL of the Job
            build_number : Build number for the Job
            latest       : Latest build

        Returns:
            Stage status text
        """
        if build_url:
            build_url = utility.build_url_complete(build_url)

        # Get the stage info
        stage_info = self.info(stage_name=stage_name,
                               build_url=build_url,
                               job_name=job_name,
                               job_url=job_url,
                               build_number=build_number,
                               latest=latest)

        if not stage_info:
            return StageStatus.NOT_FOUND.value

        # Check if in process (build is there but results not posted)
        if not stage_info['status']:
            logger.debug('Stage was found running/building, however no results are posted')
            return StageStatus.UNKNOWN.value
        else:
            logger.debug('Stage found, but has concluded or stopped with result')
            return stage_info['status']

    def step_list(self,
                  stage_name: str,
                  build_url: str = '',
                  job_name: str = '',
                  job_url: str = '',
                  build_number: Union[int, None] = None,
                  latest: bool = False) -> Tuple[list, list]:
        """List of steps for this stage

        Details: Ways of specifying the build:
            - Build URL only
            - Job URL only
            - Job name with build number or latest flag

        Args:
            stage_name   : Name of the stage
            build_url    : Direct URL of the build
            job_name     : Name of the Job
            job_url      : URL of the Job
            build_number : Build number for the Job
            latest       : Latest build

        Returns:
            List of steps, information and URL list
        """
        if build_url:
            build_url = utility.build_url_complete(build_url)

        # Getting the stage info
        stage_info = self.info(stage_name=stage_name,
                               build_url=build_url,
                               job_name=job_name,
                               job_url=job_url,
                               build_number=build_number,
                               latest=latest)

        # Check if there are steps in this stage
        if "stageFlowNodes" not in stage_info:
            fail_out('Failed to get stage step information. No stage steps listed')

        # Accounting for no stage step command
        for step_info in stage_info['stageFlowNodes']:
            if not 'parameterDescription' in step_info:
                step_info['parameterDescription'] = "No command parameters listed"

        # Getting the stage items, and getting only the names/labels of the stages
        try:
            step_list = stage_info['stageFlowNodes']
            step_name_list = [s['name'] for s in step_list]
        except KeyError as error:
            fail_out(f'Failed to parse stage information. Specific keys not found: {error}')

        return step_list, step_name_list

    def _thread_step_info(self, step_index: int, total_steps: int, step: dict) -> None:
        """TODO

        Details: TODO

        Args:
            TODO

        Returns: None
        """
        logger.debug(f'Thread starting - Step Info - (ID: {threading.get_ident()} - INDEX: {step_index}) ...')

        # Getting step information
        return_content = self.step.info(step_url=step['url_log'])

        logger.debug(f"---> {step_index+1}/{total_steps} - {step['name']}")
        if 'parameterDescription' in step:
            parameter = step['parameterDescription']
        else:
            parameter = 'None'

        # Check if there is any log text to this stage step
        if 'text' in return_content or not return_content['length'] == 0:
            # Clean up all HTML tags from return, keep only raw text
            log_text = utility.html_clean(return_content['text'])

            # Also convert to list
            log_list = [y for y in (x.strip() for x in log_text.splitlines()) if y]

            # Add extra step info to each line of log
            log_list = [f"[STEP: {step_index+1}/{total_steps}] " + s for s in log_list]

            # Add intro to the logs of this step
            log_list.insert(0,
                            f"[STEP: {step_index+1}/{total_steps}] [STEP] : {step['name']} - PARAMETER: {parameter}")
        else:
            # If no logs in step, still add step command
            log_list = [f"[STEP: {step_index+1}/{total_steps}] [STEP] : {step['name']} - PARAMETER: {parameter}"]

        with self._stage_log_list_thread_lock:
            self.stage_log_dict[step_index] = log_list

        logger.debug(f'Thread stopped - Step Info - (ID: {threading.get_ident()} - INDEX: {step_index}) ...')

    def logs(self,
             stage_name: str,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: Union[int, None] = None,
             latest: bool = False,
             download_dir: bool = False) -> bool:
        """Prints out the console log for this specified stage

        Details: Ways of specifying the build:
            - Build URL only
            - Job URL only
            - Job name with build number or latest flag

        Args:
            stage_name   : Name of the stage
            build_url    : Direct URL of the build
            job_name     : Name of the Job
            job_url      : URL of the Job
            build_number : Build number for the Job
            latest       : Latest build
            download_dir : When specified, will download logs to file in this direcotry

        Returns:
            True if success, else False
        """
        if build_url:
            build_url = utility.build_url_complete(build_url)

        # Getting all stage step information
        stage_step_list = self.step_list(stage_name=stage_name,
                                         build_url=build_url,
                                         job_name=job_name,
                                         job_url=job_url,
                                         build_number=build_number,
                                         latest=latest)[0]

        logger.debug(
            f'Downloading logs for {len(stage_step_list)} step in the stage using {len(stage_step_list)} threads ...')
        self.stage_log_dict = {}
        threads = []
        for i, stage_step in enumerate(stage_step_list):
            thread = threading.Thread(target=self._thread_step_info,
                                      args=(
                                          i,
                                          len(stage_step_list),
                                          stage_step,
                                      ),
                                      daemon=False)
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

        # Convert dict to index sorted list
        stage_log_list = []
        for _, stage_logs in sorted(self.stage_log_dict.items()):
            stage_log_list.extend(stage_logs)

        # Make the list to continuos step, with newline in between them
        stage_log_text = os.linesep.join(stage_log_list)

        if download_dir:
            # Save logs to local file
            filename = f'build-logs_{datetime.now().strftime("%m-%d-%Y_%I-%M-%S")}{self.build_logs_extension}'
            logger.debug(f'Saving console text logs to local file "{filename}" ...')
            try:
                with open(os.path.join(download_dir, filename), 'w+') as file:
                    file.write(stage_log_text)
                logger.debug('Successfully write build logs to file')
            except (IOError, PermissionError) as error:
                fail_out(f'Failed to write logs to file. Exception: {error}')
        else:
            # Output to console
            logger.debug('Printing out console text logs ...')
            print2(stage_log_text)
        stage_log_text = None

        return True



"""Tools click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import tools
from yojenkins.cli import cli_decorators, cli_tools
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@tools.command(short_help='\tOpen browser to the documentation page')
@cli_decorators.debug
def docs(debug):
    """Open browser to the documentation"""
    set_debug_log_level(debug)
    cli_tools.documentation()


# @tools.command(short_help='\tUpgrade yojenkins')
# @cli_decorators.debug
# @click.option('--user', type=bool, required=False, is_flag=True, help='Install to the Python user install directory for your platform')
# @click.option('--proxy', type=str, required=False, help='Specify a proxy in the form [user:passwd@]proxy.server:port')
# def upgrade(debug, user, proxy):
#     """Install the latest version of yojenkins. This is a thin wrapper to 'pip install'"""
#     set_debug_log_level(debug)
#     cli_tools.upgrade(user, proxy)

# @tools.command(short_help='\tRemove yojenkins')
# @cli_decorators.debug
# def remove(debug):
#     """Uninstall yojenkins using pip"""
#     set_debug_log_level(debug)
#     cli_tools.remove()


@tools.command(short_help='\tReport a bug')
@cli_decorators.debug
def bug_report(debug):
    """This command will open a web browser to report a bug"""
    set_debug_log_level(debug)
    cli_tools.bug_report()


@tools.command(short_help='\tRequest a feature')
@cli_decorators.debug
def feature_request(debug):
    """This command will open a web browser to request a feature"""
    set_debug_log_level(debug)
    cli_tools.feature_request()


@tools.command(short_help='\tShow detailed command usage history')
@cli_decorators.debug
@click.option('--profile', type=str, required=False, is_flag=False, help='Filter by profile name')
@click.option('--clear', type=bool, required=False, default=False, is_flag=True, help='Clear the history file')
def history(debug, **kwargs):
    """Show detailed command usage history"""
    set_debug_log_level(debug)
    cli_tools.history(**kwargs)


@tools.command(short_help='\tSend a generic Rest request to server')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('request_text', nargs=1, type=str, required=True)
@click.option('--request-type',
              type=click.Choice(['GET', 'POST', 'HEAD'], case_sensitive=False),
              default='GET',
              show_default=True,
              required=False,
              help='Type of Rest request')
@click.option('--raw', type=bool, required=False, default=False, is_flag=True, help='Return raw return text')
@click.option('--clean-html',
              type=bool,
              required=False,
              default=False,
              is_flag=True,
              help='Clean any HTML tags from return')
def rest_request(debug, **kwargs):
    """Use this command to send Rest calls to the server.
    The request will be send with the proper authentication from your profile.
    This can be useful if yojenkins does not have the functionality you need.

    EXAMPLE:

      - yojenkins tools rest-request "me/api/json"
    """
    set_debug_log_level(debug)
    cli_tools.rest_request(**kwargs)


@tools.command(short_help='\tRun Groovy script on server, return result')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--text', type=str, required=False, help='Command(s) to run entered as text')
@click.option('--file',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              is_flag=False,
              help='File containing command(s) to run')
@click.option('--output',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              is_flag=False,
              help='Save the result to this file')
@click.pass_context
def run_script(ctx, debug, **kwargs):
    """Use this command to execute a Groovy script, as text or in a file,
    on the Jenkins server and return the output

    EXAMPLES:

    \b
      - yojenkins tools script --text "println('hello you')"
      - yojenkins tools script --file ./my_script.groovy
    """
    set_debug_log_level(debug)
    if kwargs.get("text") or kwargs.get("file"):
        cli_tools.run_script(**translate_kwargs(kwargs))
    else:
        click.echo(ctx.get_help())


@tools.command(short_help='\tSet up a Jenkins shared library')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--lib-name', type=str, required=True, help='Name of library, to be used in the @Library annotation')
@click.option('--repo-owner', type=str, required=False, help='Owner/Organization of git repository')
@click.option('--repo-name', type=str, required=False, help='Name of git repository')
@click.option('--repo-url', type=str, required=False, help='URL of git repository. Same syntax as git clone command')
@click.option('--repo-branch',
              type=str,
              required=False,
              default='main',
              show_default=True,
              help='Branch of the git repository')
@click.option('--implicit',
              type=bool,
              required=False,
              default=False,
              show_default=True,
              is_flag=True,
              help='Automatically allow pipelines to use libraries without @Library')
@click.option('--credential-id',
              type=str,
              required=False,
              help='ID of your git credentials in Jenkins credentials database')
def shared_lib_setup(debug, **kwargs):
    """This sets up the Jenkins Shared Library, referencing a GitHub git repository.

    WARNING:

        Sharable libraries available to any Pipeline jobs running on this system.
        These libraries will be fully trusted, meaning theyrun without “sandbox” restrictions and may use @Grab.

    USAGE NOTES:

    \b
        - As of now, only GitHub repositories are supported
        - Use with --repo-owner and --repo-name [OR] --repo-url
        - Using the same --lib-name will overwrite currently defined library

    EXAMPLE:

    \b
        yojenkins tools shared-lib-setup \\
            --lib-name SHARED-LIB-NAME \\
            --repo-url https://github.com/ORG/REPO-NAME.git \\
            --repo-branch main \\
            --implicit \\
            --credential-id my-jenkins-cred-id


    """
    set_debug_log_level(debug)
    cli_tools.shared_lib_setup(**translate_kwargs(kwargs))



"""Auth Menu CLI Entrypoints"""

import logging

import click

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history
from yojenkins.yo_jenkins import Auth, Rest

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def configure(auth_file: str) -> None:
    """Configure yojenkins authentication profile(s)

    Args:
        auth_file: (Optional) Path to the the authentication setup JSON file
    """
    Auth().configure(auth_file)
    click.secho('Successfully configured credentials file', fg='bright_green', bold=True)


# @log_to_history
# def remove(profile: str) -> None:
#     """Remove yojenkins authentication profile(s)
#
#     Args:
#         profile: The yojenkins profile name to remove from credentials
#     """
#     Auth().remove(profile_name=profile)
#     click.secho('Successfully removed credentials profile', fg='bright_green', bold=True)


@log_to_history
def token(profile: str, token: str, name: str, server_base_url: str, username: str, password: str) -> None:
    """Generate authentication API token

    Args:
        profile: The profile/account to use
        token:   API Token for Jenkins server
        name: Name of the generated token
        server_base_url: Server base URL address
        username: Account username
        password: User password to use to generate API token
    """
    auth = Auth()

    if profile:
        # Add/Refresh the newly generated API token for an existing profile
        data = auth.profile_add_new_token(profile_name=profile, token=token, password=password)
    else:
        # Simply display/output the newly generated API Token
        if token:
            click.secho('Ignoring --token. Must be specfied with --profile', fg="yellow")
            print()
        data = auth.generate_token(token_name=name,
                                   server_base_url=server_base_url,
                                   username=username,
                                   password=password)
    if profile:
        click.secho('success', fg='bright_green', bold=True)
    else:
        click.secho(data, fg='bright_green', bold=True)


@log_to_history
def show(**kwargs) -> None:
    """Show the local credentials profiles

    Args:
        kwargs: Various CLI options
    """
    data = Auth().show_local_credentials()
    cu.standard_out(data, **kwargs)


@log_to_history
def verify(profile: str) -> None:
    """Check if credentials can authenticate

    Args:
        profile: The profile/account to use
    """
    auth = Auth(Rest())
    auth.get_credentials(profile)
    auth.create_auth()
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def user(profile: str, token: str, **kwargs) -> None:
    """Show current user information

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
    """
    data = cu.config_yo_jenkins(profile, token).auth.user()
    cu.standard_out(data, **kwargs)



"""This is just a set of dummy tests"""

import logging


def test_success() -> None:
    """This is just a dummy test"""
    logging.info("This is just a dummy test that SUCCEEDS")
    assert True


def test_failure() -> None:
    """This is just a dummy test"""
    logging.info("This is just a dummy test that FAILS")
    assert False



"""Containerized Jenkins instance Management"""

import logging
import os
import platform
import shutil
from datetime import datetime

if platform.system() != "Windows":
    # Non-windows systems get group ID
    from grp import getgrnam

from time import perf_counter
from typing import Any, Dict, Tuple

import docker
from docker.errors import DockerException

from yojenkins.utility.utility import fail_out, get_resource_path, print2

# Getting the logger reference
logger = logging.getLogger()


class DockerJenkinsServer():
    """Class managing containerized Jenkins instance"""

    def __init__(self,
                 config_file: str = 'config_as_code.yaml',
                 plugins_file: str = 'plugins.txt',
                 protocol_schema: str = 'http',
                 host: str = 'localhost',
                 port: int = 8080,
                 image_fullname: str = 'yojenkins-jenkins:latest',
                 image_base: str = 'jenkins/jenkins',
                 extra_setup_script: str = '',
                 image_rebuild: bool = False,
                 new_volume: bool = False,
                 new_volume_name: str = 'yojenkins-jenkins',
                 bind_mount_dir: str = '',
                 container_name: str = 'yojenkins-jenkins',
                 registry: str = '',
                 admin_user: str = 'admin',
                 password: str = 'password'):
        """Object constructor method, called at object creation

        ### TODO: Use kwargs instead of positional arguments

        Args:
            None

        Returns:
            None
        """

        # TODO: Option flag to mount docker socket (for security)

        self.docker_client = None

        # Image Related
        self.docker_registry = registry
        self.image_base_image = image_base
        self.image_base_version = 'latest'
        self.image_dockerfile_dir = get_resource_path(os.path.join('resources', 'server_docker_settings'))
        self.image_fullname = image_fullname
        self.image_rebuild = image_rebuild
        self.image_build_args = {
            "JENKINS_BASE_IMAGE": self.image_base_image,
            "JENKINS_BASE_VERSION": self.image_base_version,
            "JENKINS_CONFIG_FILE": config_file,
            "JENKINS_PLUGINS_FILE": plugins_file,
            "PROTOCOL_SCHEMA": protocol_schema,
            "JENKINS_HOSTNAME": host,
            "JENKINS_PORT": f"{port}",
            "JENKINS_ADMIN_ID": admin_user,
            "JENKINS_ADMIN_PASSWORD": password
        }

        # Container Related
        self.container_name = container_name
        self.container_address = f"{protocol_schema}://{host}:{port}"
        self.container_ports = {'8080/tcp': port, '50000/tcp': 50000}
        self.container_env_vars = []
        self.container_restart_policy = {}
        self.container_remove = False if self.container_restart_policy else True

        # Volume Related
        self.volumes_bind = {} if not bind_mount_dir else {bind_mount_dir: {'bind': '/tmp/my_things', 'mode': 'rw'}}
        self.volumes_named = {f'{new_volume_name}': '/var/jenkins_home'}
        self.volumes_mounts = []
        self.new_volume = new_volume

        # Extra/Custom Setup Script File
        self.extra_setup_script = extra_setup_script
        if extra_setup_script:
            logger.debug("Extra setup script for image build provided")
            logger.debug(f'Copying extra setup script to Docker context directory ...')
            source_path = os.path.abspath(extra_setup_script)
            target_path = os.path.join(self.image_dockerfile_dir, 'extra_setup_script.sh')
        else:
            logger.debug("Extra setup script for image build NOT provided")
            logger.debug(f'Copying empty dummy script as extra setup script ...')
            source_path = os.path.join(self.image_dockerfile_dir, 'dummy.sh')
            target_path = os.path.join(self.image_dockerfile_dir, 'extra_setup_script.sh')
        logger.debug(f'    - Source: {source_path}')
        logger.debug(f'    - Target: {target_path}')
        try:
            shutil.copy2(source_path, target_path)
        except (IOError, PermissionError) as error:
            fail_out(f'Failed to copy "{source_path}" to "{target_path}". Exception: {error}')

        # Output all attributes
        logger.debug(f'Jenkins server build attributes:')
        for key, value in vars(self).items():
            if key in ['image_build_args', 'docker_client', 'container_env_vars']:
                continue
            logger.debug(f'    - {key}: {value}')

    def docker_client_init(self) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug('Connecting to local docker client/engine ...')
        # Get the local docker client
        try:
            self.docker_client = docker.from_env()
        except DockerException as error:
            logger.debug(f'Failed to get docker client/engine handle. Exception: {error}')
            return False

        # Ping the docker client
        if not self.docker_client.ping():
            logger.debug('Failed to ping local docker client')
            return False
        logger.debug('Successfully connected to local docker client/engine')

        return True

    def setup(self) -> dict:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        deployed: Dict[str, Any] = {}

        if not self.docker_client:
            if not self.docker_client_init():
                logger.debug('Docker client was not initialized. Please run .docker_client_init() first')
                return deployed

        if not self._container_kill():
            return deployed, False
        if self.image_rebuild:
            if not self._image_remove():
                return deployed, False

        image_name = self._image_build()
        if not image_name:
            return deployed, False
        deployed['image'] = image_name

        volume_names = self._volumes_create()
        if not volume_names:
            return deployed, False
        deployed['volumes'] = volume_names

        container_name, server_address = self._container_run()
        if not server_address:
            return deployed, False
        deployed['container'] = container_name
        deployed['address'] = server_address
        deployed['deploy_timestamp'] = datetime.now().timestamp()
        deployed['deploy_datetime'] = datetime.now().strftime("%A, %B %d, %Y %I:%M:%S")
        deployed['docker_version'] = self.docker_client.info(
        )['ServerVersion'] if 'ServerVersion' in self.docker_client.info() else 'N/A'

        return deployed, True

    def clean(self, remove_volume: bool = False, remove_image: bool = False) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        removed = {}

        if not self._container_kill():
            return False

        if remove_volume:
            if not self._volumes_remove():
                return False

        if remove_image:
            if not self._image_remove():
                return False

        return True

    def _image_build(self) -> str:
        """Building docker image with specified parameters

        Details:
            - https://docker-py.readthedocs.io/en/stable/images.html

        Returns:
            Name of the image that was built
        """
        start_time = perf_counter()
        logger.debug(f'Building image: {self.image_fullname} ...')
        logger.debug(f'Dockerfile context directory: {self.image_dockerfile_dir}')
        try:
            _, build_logs = self.docker_client.images.build(path=self.image_dockerfile_dir,
                                                            tag=self.image_fullname,
                                                            rm=True,
                                                            buildargs=self.image_build_args,
                                                            quiet=False,
                                                            forcerm=True)
            logger.debug('Docker image build logs:')
            for chunk in build_logs:
                if 'stream' in chunk:
                    for line in chunk['stream'].splitlines():
                        logger.debug(f"    {line}")
        except DockerException as error:
            print2(f'Failed to build image: {self.image_fullname}. Exception: {error}', bold=True, color="red")
            if ": 1" in str(error):
                print2('   Error Code 1   : Operation not permitted', bold=True, color="red")
                print2('   Possible Reason: General error script', bold=True, color="red")
            elif ": 127" in str(error):
                print2('   Error Code 127 : Shell command not found', bold=True, color="red")
                print2('   Possible Reason: Unknown command, misspelling, or end of line sequence (CRLF/LF)',
                       bold=True,
                       color="red")
            elif ": 2" in str(error):
                print2('   Error Code 2   : No such file exists', bold=True, color="red")
                print2('   Possible Reason: Missing script or command references missing file', bold=True, color="red")
            elif ": 126" in str(error):
                print2('   Error Code 126 : Permission Issue', bold=True, color="red")
                print2('   Possible Reason: Script or command does not have sufficient privileges',
                       bold=True,
                       color="red")
            else:
                print2('   Please see error code: https://www.thegeekstuff.com/2010/10/linux-error-codes/',
                       bold=True,
                       color="red")
            return ''
        logger.debug(
            f'Successfully build image: {self.image_fullname} (Elapsed time: {perf_counter() - start_time:.3f}s)')
        return self.image_fullname

    def _image_remove(self) -> bool:
        """Remove a specified docker image

        Returns:
            True if successful, False otherwise
        """
        logger.debug(f'Removing image: {self.image_fullname} ...')
        try:
            self.docker_client.images.remove(self.image_fullname)
        except DockerException as error:
            logger.debug(f'Failed to remove image: {self.image_fullname}. Exception: {error}')
            return False
        logger.debug(f'Successfully removed image: {self.image_fullname}')
        return True

    def _volumes_create(self) -> list:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        self.volumes_mounts = []
        volume_mounts_names = []

        # Bind Volume Mounts
        for volume_source, volume_target in self.volumes_bind.items():
            logger.debug(f'Adding new local bind volume to mount list: {volume_source} -> {volume_target} ...')
            mount = docker.types.Mount(source=volume_source, target=volume_target, type='bind', read_only=False)
            self.volumes_mounts.append(mount)
            volume_mounts_names.append({'bind': volume_source})

        # Named Volume Mounts
        for volume_name, volume_dir in self.volumes_named.items():
            try:
                volume_handle = self.docker_client.volumes.get(volume_name)
                if self.new_volume:
                    logger.debug(f'Removing found matching/existing named volume: {volume_name} ...')
                    volume_handle.remove(force=True)
                else:
                    logger.debug(f'Using found matching/existing named volume: {volume_name}')
            except:
                logger.debug(f'Creating new named volume: {volume_name}')
                self.docker_client.volumes.create(name=volume_name, driver='local')
                logger.debug('Successfully created!')

            logger.debug(f'Adding named volumes to mount list: {volume_name} ...')
            mount = docker.types.Mount(target=volume_dir, source=volume_name, type='volume', read_only=False)
            self.volumes_mounts.append(mount)
            volume_mounts_names.append({'named': volume_name})

        # Adding Docker unix socket
        logger.debug('Adding bind mount to Docker unix socket: /var/run/docker.sock ...')
        mount = docker.types.Mount(target='/var/run/docker.sock',
                                   source='/var/run/docker.sock',
                                   type='bind',
                                   read_only=False)
        self.volumes_mounts.append(mount)

        logger.debug(f'Number of volumes to be mounted to container: {len(self.volumes_mounts)}')
        return volume_mounts_names

    def _volumes_remove(self) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Removing named volumes: {self.volumes_named.keys()}')
        for volume_name in self.volumes_named:
            try:
                volume_named = self.docker_client.volumes.get(volume_name)
                volume_named.remove(force=True)
                logger.debug(f'    - Volume: {volume_name} - REMOVED')
            except DockerException as error:
                logger.debug(f'    - Volume: {volume_name} - FAILED - Exception: {error}')
        return True

    def _container_run(self) -> Tuple[str, str]:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        # Getting docker group id (Unix only)
        if platform.system() != "Windows":
            docker_gid = [getgrnam('docker').gr_gid]
        else:
            docker_gid = None

        logger.debug(f'Local docker service group id found: {docker_gid}')

        logger.debug(f'Creating and running container: {self.container_name} ...')
        try:
            self.docker_client.containers.run(name=self.container_name,
                                              image=self.image_fullname,
                                              environment=self.container_env_vars,
                                              ports=self.container_ports,
                                              mounts=self.volumes_mounts,
                                              restart_policy=self.container_restart_policy,
                                              remove=self.container_remove,
                                              auto_remove=self.container_remove,
                                              detach=True,
                                              group_add=docker_gid)
        except DockerException as error:
            logger.debug(f'Failed to run container: {self.container_name} Exception: {error}')
            return '', ''
        logger.debug(f'Successfully running container: {self.container_name}')
        logger.debug(f'Container "{self.container_name}" is running at this address:')
        logger.debug(f'    --> {self.container_address}')
        return self.container_name, self.container_address

    def _container_stop(self) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Stopping container with matching tag: {self.container_name} ...')
        try:
            container_handle = self.docker_client.containers.get(self.container_name)
            container_handle.stop()
            logger.debug('Container is stopped')
        except DockerException as error:
            logger.debug(f'Failed to stop container matching tag: {self.container_name}. Exception: {error}')
            return False
        return True

    def _container_kill(self) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Killing container with matching tag: {self.container_name} ...')
        try:
            container_handle = self.docker_client.containers.get(self.container_name)
            container_handle.kill()
            logger.debug('Container is dead')
        except DockerException as error:
            logger.debug(f'Failed to kill container matching tag: {self.container_name}. Exception: {error}')
        return True



"""Credential Menu CLI Entrypoints"""

import json
import logging

import click
import xmltodict

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def list(profile: str, token: str, opt_list: bool, folder: str, domain: str, keys: str, **kwargs) -> None:
    """List credentials

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        folder: The folder to use for the credential
        domain: The domain to use for the credential
        keys: Credential keys to list
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data, data_list = yj_obj.credential.list(folder=folder, domain=domain, keys=keys)
    data = data_list if opt_list else data
    cu.standard_out(data, **kwargs)


@log_to_history
def info(profile: str, token: str, credential: str, folder: str, domain: str, **kwargs) -> None:
    """Credential information

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        folder: The folder to use for the credential
        domain: The domain to use for the credential
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data = yj_obj.credential.info(credential=credential, folder=folder, domain=domain)
    cu.standard_out(data, **kwargs)


@log_to_history
def config(profile: str, token: str, opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, opt_json: bool,
           credential: str, folder: str, domain: str, filepath: str) -> None:
    """Get credential configuration

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        folder: The folder to use for the credential
        domain: The domain to use for the credential
        filepath: The filepath of the credential configuration
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data = yj_obj.credential.config(credential=credential,
                                    folder=folder,
                                    domain=domain,
                                    opt_json=opt_json,
                                    opt_yaml=opt_yaml,
                                    opt_toml=opt_toml,
                                    filepath=filepath)
    opt_xml = not any([opt_json, opt_yaml, opt_toml])
    data = data if opt_xml else json.loads(json.dumps(xmltodict.parse(data)))
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def get_template(profile: str, token: str, opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool,
                 opt_json: bool, type: str, filepath: str) -> None:
    """Credential type template to create a credential

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        cred_type: The credential type to get the template for
        filepath: The filepath to save the credential configuration
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    data = yj_obj.credential.get_template(cred_type=type,
                                          opt_json=opt_json,
                                          opt_yaml=opt_yaml,
                                          opt_toml=opt_toml,
                                          filepath=filepath)
    opt_xml = not any([opt_json, opt_yaml, opt_toml])
    data = data if opt_xml else json.loads(json.dumps(xmltodict.parse(data)))
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def create(profile: str, token: str, config_file: str, folder: str, domain: str) -> None:
    """Create new credentials

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        config_file: The filepath of the credential configuration
        folder: The folder to use for the credential
        domain: The domain to use for the credential
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.credential.create(config_file=config_file, folder=folder, domain=domain)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def delete(profile: str, token: str, credential: str, folder: str, domain: str) -> None:
    """Remove credentials

    Args:
        profile: The profile/account to use
        token:   API token for Jenkins server
        credential: The credential ID or URL to delete
        folder: The folder of the credential
        domain: The domain of the credential
    """
    yj_obj = cu.config_yo_jenkins(profile, token)
    yj_obj.credential.delete(credential=credential, folder=folder, domain=domain)
    click.secho('success', fg='bright_green', bold=True)



"""Folder click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import folder
from yojenkins.cli import cli_decorators, cli_folder
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@folder.command(short_help='\tFolder information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
def info(debug, **kwargs):
    """Folder information"""
    set_debug_log_level(debug)
    cli_folder.info(**translate_kwargs(kwargs))


@folder.command(short_help='\tSearch folders by REGEX pattern')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('search_pattern', nargs=1, type=str, required=True)
@click.option('-sf', '--search-folder', type=str, default='', required=False, help='Folder within which to search')
@click.option('-d', '--depth', type=int, default=4, required=False, help='Search depth from root directory')
@click.option('-fn',
              '--fullname',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Search entire folder path name')
@cli_decorators.list
def search(debug, **kwargs):
    """Search folders by REGEX pattern"""
    set_debug_log_level(debug)
    cli_folder.search(**translate_kwargs(kwargs))


@folder.command(short_help='\tList all subfolders in folder')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@cli_decorators.list
def subfolders(debug, **kwargs):
    """List all subfolders in folder"""
    set_debug_log_level(debug)
    cli_folder.subfolders(**translate_kwargs(kwargs))


@folder.command(short_help='\tList all jobs in folder')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@cli_decorators.list
def jobs(debug, **kwargs):
    """List all jobs in folder"""
    set_debug_log_level(debug)
    cli_folder.jobs(**translate_kwargs(kwargs))


@folder.command(short_help='\tList all views in folder')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@cli_decorators.list
def views(debug, **kwargs):
    """List all views in folder"""
    set_debug_log_level(debug)
    cli_folder.views(**translate_kwargs(kwargs))


@folder.command(short_help='\tList all items in folder')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@cli_decorators.list
def items(debug, **kwargs):
    """List all items in folder"""
    set_debug_log_level(debug)
    cli_folder.items(**translate_kwargs(kwargs))


@folder.command(short_help='\tOpen folder in web browser')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
def browser(debug, **kwargs):
    """Open folder in web browser"""
    set_debug_log_level(debug)
    cli_folder.browser(**translate_kwargs(kwargs))


@folder.command(short_help='\tGet folder configuration')
@cli_decorators.debug
@cli_decorators.format_output
@click.option('-j',
              '--json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Output config in JSON format')
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@click.option('--filepath',
              type=click.Path(file_okay=True, dir_okay=True),
              required=False,
              help='File/Filepath to write configurations to')
def config(debug, **kwargs):
    """Get folder configuration"""
    set_debug_log_level(debug)
    cli_folder.config(**translate_kwargs(kwargs))


@folder.command(short_help='\tCreate an item [folder, view, job]')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('name', nargs=1, type=str, required=True)
@click.argument('folder', nargs=1, type=str, required=True)
@click.option('--type',
              type=click.Choice(['folder', 'view', 'job'], case_sensitive=False),
              default='folder',
              show_default=True,
              required=False,
              help='Item type created')
@click.option('--config-file',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              help='Path to local file defining item')
@click.option('--config-is-json',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='The specified file is in JSON format')
def create(debug, **kwargs):
    """Create an item [folder, view, job]"""
    set_debug_log_level(debug)
    cli_folder.create(**translate_kwargs(kwargs))


@folder.command(short_help='\tCopy an existing item')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
@click.argument('original', nargs=1, type=str, required=True)
@click.argument('new', nargs=1, type=str, required=True)
def copy(debug, **kwargs):
    """Copy an existing item"""
    set_debug_log_level(debug)
    cli_folder.copy(**translate_kwargs(kwargs))


@folder.command(short_help='\tDelete folder or view')
@cli_decorators.debug
@cli_decorators.profile
@click.argument('folder', nargs=1, type=str, required=True)
def delete(debug, **kwargs):
    """Delete folder or view"""
    set_debug_log_level(debug)
    cli_folder.delete(**translate_kwargs(kwargs))



"""yojenkins package version defintion"""

__title__ = 'yojenkins'
__summary__ = "A CLI tool to interface with Jenkins server"
__version__ = "0.0.85"
__author__ = "Ismet Handzic"
__copyright__ = "Copyright 2022 %s" % __author__



"""Credential class definition"""

import json
import logging
import os
import xml.etree.ElementTree as ET
from json.decoder import JSONDecodeError
from typing import Dict, Tuple

from json2xml import json2xml
from urllib3.util import parse_url

from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out
from yojenkins.yo_jenkins.jenkins_item_template import JenkinsItemTemplate

# Getting the logger reference
logger = logging.getLogger()


class Credential():
    """Credential class"""

    def __init__(self, rest) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest

    @staticmethod
    def _get_folder_store(folder: str) -> Tuple[str, str]:
        """Utility method to get credential folder name and domain

        Args:
            folder: folder name or url

        Returns:
            Folder nae and store name
        """
        if folder and utility.is_full_url(folder):
            folder = utility.url_to_name(folder)
            store = 'folder'
            logger.debug(f'Credential folder name or url passed. Using effective store: "{store}"')
        if folder in ['root', '.']:
            folder = '.'
            store = 'system'
            logger.debug(f'Using effective credential folder name: "" = "{folder}"')
        else:
            if "job/" not in folder:
                folder = f"job/{folder}"
            store = 'folder'
        return folder, store

    @staticmethod
    def _get_domain(domain: str) -> str:
        """Utility method to get credential domain name

        Args:
            domain: Credential domain name

        Returns:
            Effective domain name
        """
        domain_effective = domain
        if domain in ["global"]:
            domain_effective = "_"
            logger.debug(f'Credential domain passed: "{domain}". Using effective domain: "{domain_effective}"')
        return domain_effective

    @staticmethod
    def _get_folder_store_domain_from_url(credential_url: str) -> Tuple[str, str, str]:
        """Utility method to get folder name, store name and domain from a credential URL

        Details:
            Example credential_url formats:
                - /job/my-folder/credentials/store/folder/domain/_/credential/<CRED-ID>/
                - /credentials/store/system/domain/_/credential/<CRED-ID>
                - /./credentials/store/folder/domain/_/credential/<CRED-ID>/

        Args:
            credential_url: Credential url

        Returns:
            Folder name, store name and domain
        """
        parsed_path = parse_url(credential_url).path.strip('/').split('/')
        key_words = ['credentials', 'store', 'domain']
        if any(x in parsed_path for x in key_words):
            # Folder
            credential_index = parsed_path.index('credentials')
            if credential_index > 2:
                logger.debug('Failed to parse the credential URL. "credentials" keyword in URL in wrong position')
                return "", "", ""
            elif credential_index == 2:
                folder = "job/" + parsed_path[credential_index - 1]
            else:
                folder = "."

            # Store
            store_index = parsed_path.index('store')
            store = parsed_path[store_index + 1]

            # Domain
            domain_index = parsed_path.index('domain')
            domain = parsed_path[domain_index + 1]
        else:
            logger.debug(
                f'The credential URL "{credential_url}" path does not contain any of the expected keywords: {key_words}'
            )
            return "", "", ""
        logger.debug(
            f'Successfully parsed the credential URL. Folder: "{folder}", Store: "{store}", Domain: "{domain}"')

        return folder, store, domain

    def list(self, domain: str, keys: str, folder: str = None) -> Tuple[list, list]:
        """List all credentials for the specified folder and domain

        Details:
            - Available credentials key may change over time
            - May use the Web UI to get a sense what the domain is for a credential

        Args:
            domain: Credential domain name
            keys: Credential keys to list
            folder: Credential folder name

        Returns:
            List of credentials in dictionary format and a list of credential names
        """
        folder, store = self._get_folder_store(folder)
        domain = self._get_domain(domain)

        # Get return keys
        if keys in ["all", "*", "full"]:
            keys = "*"
        else:
            keys = utility.parse_and_check_input_string_list(keys, ',')

        logger.debug('Getting all credentials for the following:')
        logger.debug(f'   - Folder: {folder}')
        logger.debug(f'   - Store:  {store}')
        logger.debug(f'   - Domain: {domain}')
        logger.debug(f'   - Keys:   {keys}')

        target = f'{folder}/credentials/store/{store}/domain/{domain}/api/json?tree=credentials[{keys}]'
        credentials_info, _, success = self.rest.request(target=target,
                                                         request_type='get',
                                                         is_endpoint=True,
                                                         json_content=True)
        if not success:
            fail_out('Failed to get any credentials')

        if "credentials" not in credentials_info:
            fail_out('Failed to find "credentials" section in request return content')
        credential_list = credentials_info["credentials"]
        if not any(credential_list):
            fail_out('No credentials listed for the specified folder or domain')

        # Get a list of only credentail names
        credential_list_name = [
            credential["displayName"] for credential in credential_list if "displayName" in credential
        ]

        logger.debug(f'Number of credentials found: {len(credential_list)}')
        logger.debug(f'Found the following credential names: {credential_list_name}')

        return credential_list, credential_list_name

    def info(self, credential: str, folder: str, domain: str) -> Dict:
        """Getting credential info

        Args:
            credential: credential name, url, or ID
            folder: folder name or url
            store: store name
            domain: domain name

        Returns:
            Credential inforamation in dictionary format
        """
        logger.debug(f"Getting credential info for: {credential} ...")
        is_endpoint = True

        # Check if credential is a url or name/ID
        if utility.is_full_url(credential):
            logger.debug('Using direct credential URL ...')
            target = f'{credential.strip("/")}/api/json'
            is_endpoint = False

            folder, store, domain = self._get_folder_store_domain_from_url(credential)
            credential_url = credential
        else:
            folder_original = folder
            folder, store = self._get_folder_store(folder)
            domain = self._get_domain(domain)

            credentials_list, _ = self.list(domain=domain, keys="displayName,id", folder=folder_original)
            credential_ids_match = []
            for credential_item in credentials_list:
                for key_name in ['displayName', 'id', 'fullName']:
                    if credential_item.get(key_name) == credential.lower():
                        credential_ids_match.append(credential_item["id"])
                        logger.debug(f'Successfully found credential matching '
                                     f'{key_name} "{credential}" ({credential_item["id"]})')

            if not credential_ids_match:
                fail_out(f'Failed to find any credentials matching display name or ID: {credential}')
            if len(credential_ids_match) > 1:
                logger.debug(f'More than one matching credential found. '
                             f'Using the first one: {credential_ids_match[0]}')
            target = f'{folder}/credentials/store/{store}/domain/{domain}/credential/{credential_ids_match[0]}/api/json'
            credential_url = f'{self.rest.server_url}{folder}/credentials/store/{store}/domain/{domain}/credential/{credential_ids_match[0]}'

        logger.debug('Getting all credential info with the following info:')
        logger.debug(f'   - Folder:     {folder}')
        logger.debug(f'   - Store:      {store}')
        logger.debug(f'   - Domain:     {domain}')
        logger.debug(f'   - Credential: {credential}')

        # Get credential info
        credential_info, _, success = self.rest.request(target=target,
                                                        request_type='get',
                                                        is_endpoint=is_endpoint,
                                                        json_content=True)
        if not success:
            fail_out('Failed to get credential information')

        # Add URL to the credential info
        credential_info['url'] = credential_url

        return credential_info

    def config(self,
               credential: str,
               folder: str = None,
               domain: str = None,
               filepath: str = None,
               opt_json: bool = False,
               opt_yaml: bool = False,
               opt_toml: bool = False) -> str:
        """Get the folder configuration (ie .config.xml)

        Args:
            credential: Credential name, url, or ID
            folder: Credential folder name or url
            domain: Credential domain name
            filepath: Path to the file to be written
            opt_json: Write in JSON format
            opt_yaml: Write in YAML format
            opt_toml: Write in TOML format

        Returns:
            Folder config.xml contents
        """
        folder, store = self._get_folder_store(folder)
        domain = self._get_domain(domain)

        credential_info = self.info(credential=credential, folder=folder, domain=domain)
        credential_id = credential_info.get('id')
        if not credential_id:
            fail_out('Failed to find "id" key within credential information')

        # If URL passed, parse out folder and domain
        if utility.is_full_url(credential):
            folder, store, domain = self._get_folder_store_domain_from_url(credential)

        target = f'{folder}/credentials/store/{store}/domain/{domain}/credential/{credential_id}/config.xml'
        logger.debug(f'Fetching XML configurations for credential: "{credential_id}" ...')
        return_content, _, success = self.rest.request(target=target,
                                                       request_type='get',
                                                       json_content=False,
                                                       is_endpoint=True)
        logger.debug('Successfully fetched XML configurations' if success else 'Failed to fetch XML configurations')

        if filepath:
            write_success = utility.write_xml_to_file(return_content, filepath, opt_json, opt_yaml, opt_toml)
            if not write_success:
                fail_out('Failed to write configuration file')

        return return_content

    def get_template(self,
                     cred_type: str,
                     filepath: str = None,
                     opt_json: bool = False,
                     opt_yaml: bool = False,
                     opt_toml: bool = False) -> str:
        """Get the template file for a credential type

        Args:
            cred_type: Credential type
            filepath: Path to the template to be written
            opt_json: Write in JSON format
            opt_yaml: Write in YAML format
            opt_toml: Write in TOML format

        Returns:
            Template string and file write success
        """
        # Get the template file
        template = JenkinsItemTemplate.CREDENTIAL.value[cred_type]
        template = template.strip(os.linesep)

        if filepath:
            write_success = utility.write_xml_to_file(template, filepath, opt_json, opt_yaml, opt_toml)
            if not write_success:
                fail_out('Failed to write configuration file')

        return template

    def create(self, config_file: str, folder: str, domain: str) -> bool:
        """Create a credential

        Args:
            config_file: Local path to file of the credential configuration
            folder: folder name or url
            store: store name
            domain: domain name

        Returns:
            True if credential created, else False
        """
        logger.debug(f'Opening and reading file: {config_file} ...')
        try:
            with open(config_file, 'rb') as open_file:
                credential_config = open_file.read()
        except (OSError, IOError, PermissionError) as error:
            fail_out(f'Failed to open and read file: {config_file}  Exception: {error}')

        try:
            cred_config_dict = json.loads(credential_config)
            logger.debug('Configuration file passed is in JSON format')
            logger.debug('Converting JSON file to XML format ...')

            # Convert to XML, remove root XML tag, and converting to string
            credential_config_xml = json2xml.Json2xml(cred_config_dict, pretty=False, wrapper="root",
                                                      attr_type=False).to_xml()
            credential_config_xml = list(ET.fromstring(credential_config_xml))[0]
            credential_config_xml = ET.tostring(credential_config_xml,
                                                encoding='utf8',
                                                method='xml',
                                                xml_declaration=False)
        except JSONDecodeError as error:
            logger.debug('Configuration file passed is in XML format')
            credential_config_xml = credential_config

        folder, store = self._get_folder_store(folder)
        domain = self._get_domain(domain)

        # # TEMPLATE WORK
        # test = utility.template_apply(string_template=CRED_USER_PASS, is_json=True,
        #                             domain=domain,
        #                             username='some-username',
        #                             password='MYPASSWORD',
        #                             description=description)
        # print(test)

        target = f'{folder}/credentials/store/{store}/domain/{domain}/createCredentials'
        _, _, success = self.rest.request(target=target,
                                          request_type='post',
                                          json_content=False,
                                          is_endpoint=True,
                                          headers={'Content-Type': 'application/xml; charset=utf-8'},
                                          data=credential_config_xml)
        if not success:
            fail_out('Failed to create credential')
        logger.debug('Successfully created credential')

        return success

    def delete(self, credential: str, folder: str = None, domain: str = None) -> bool:
        """Delete a credential

        Args:
            credential: Credential name, url, or ID
            folder: Credential folder name or url
            domain: Credential domain name

        Returns:
            True if credential deleted, else False
        """
        folder, store = self._get_folder_store(folder)
        domain = self._get_domain(domain)

        credential_info = self.info(credential=credential, folder=folder, domain=domain)
        if not credential_info:
            fail_out('Failed to get credential information. No folder name or folder url received')

        credential_id = credential_info.get('id')
        if not credential_id:
            fail_out('Failed to find "id" key within credential information')

        target = f'{folder}/credentials/store/{store}/domain/{domain}/credential/{credential_id}/config.xml'
        _, _, success = self.rest.request(target=target, request_type='delete', is_endpoint=True)
        if not success:
            fail_out('Failed to delete credential')
        logger.debug('Successfully deleted credential')

        return success



"""Step click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import step
from yojenkins.cli import cli_decorators, cli_step
from yojenkins.cli.cli_utility import set_debug_log_level
from yojenkins.utility.utility import translate_kwargs


@step.command(short_help='\tStep information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@click.argument('url', nargs=1, type=str, required=True)
def info(debug, **kwargs):
    """Step information"""
    set_debug_log_level(debug)
    cli_step.info(**translate_kwargs(kwargs))



"""General utility and tools"""

import difflib
import json
import logging
import os
import re
import sys
import sysconfig
import time
import webbrowser
from pathlib import Path
from string import Template
from typing import Any, Dict, List, Set, Tuple, Union
from urllib.parse import urljoin, urlparse

import requests
import toml
import xmltodict
import yaml
from click import echo, secho, style
from urllib3.util import parse_url
from yaspin import yaspin
from yaspin.spinners import Spinners

from yojenkins import __version__
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses

logger = logging.getLogger()

CONFIG_DIR_NAME = ".yojenkins"
KWARG_TRANSLATE_MAP = {
    "pretty": "opt_pretty",
    "yaml": "opt_yaml",
    "xml": "opt_xml",
    "toml": "opt_toml",
    "list": "opt_list",
    "json": "opt_json",
    "id": "opt_id",
}


class TextStyle:
    """Text style definitions"""

    HEADER = "\033[95m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    NORMAL = "\033[0m"


def translate_kwargs(original_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Rename the key names of the provided kwargs based on
    a set of translation rules.

    Details: This is primarirly used if kwarg keys are named like
             a module or python native object.
             Examples:
                - pretty -> opt_pretty
                - json -> opt_json

    Args:
        original_kwargs: Key-value data to be updated

    Returns:
        Updated kwargs
    """
    new_kwargs = {}
    for key, value in original_kwargs.items():
        if key in list(KWARG_TRANSLATE_MAP.keys()):
            new_kwargs[KWARG_TRANSLATE_MAP[key]] = value
        else:
            new_kwargs[key] = value
    return new_kwargs


def print2(message: str, bold: bool = False, color: str = "reset") -> None:
    """Print a message to the console using click

    Details:
        - Colors: `black` (might be a gray), `red`, `green`, `yellow` (might be an orange), `blue`,
          `magenta`, `cyan`, `white` (might be light gray), `reset` (reset the color code only)

    Example Usage:
        - `print2("Hey there!", bold=True, color="green")`

    Args:
        message: Message to print to console
        bold   : Whether to bold the message
        color  : Color to use for the message ()
    """
    echo(style(message, fg=color, bold=bold))


def fail_out(message: str) -> None:
    """Output one failure message to the console, then exit

    Example Usage:
        - `fail_out("Something went wrong!")`

    Args:
        message: Message to output to console
    """
    echo(style(message, fg="bright_red", bold=True))
    sys.exit(1)


def failures_out(messages: list) -> None:
    """Output multiple failure messages to the console, then exit

    Example Usage:
        - `failures_out(["Oh no!", "This is not good!"])`

    Args:
        message: Multiple messages to output to console
    """
    for message in messages:
        echo(style(message, fg="bright_red", bold=True))
    sys.exit(1)


def load_contents_from_local_file(file_type: str, local_file_path: str) -> Dict:
    """Loading a local file contents

    ### TODO: Add default option to load file as plain text

    Parameters:
        file_type (str)       : Type of file to be loaded ie. 'yaml', 'toml', 'json'
        local_file_path (str) : Path to a local TOML file to be loaded
    Returns:
        file_contents (dict) : The contents of file
    """
    file_type = file_type.lower()

    # Check if file exists
    if not os.path.isfile(local_file_path):
        fail_out(f"Failed to find file: {local_file_path}")

    # Check if file is completely empty
    if os.stat(local_file_path).st_size == 0:
        return {}

    logger.debug(f"Loading specified local .{file_type} file: '{local_file_path}' ...")
    try:
        with open(local_file_path, "r") as open_file:
            if file_type == "yaml":
                file_contents = yaml.safe_load(open_file)
            elif file_type == "toml":
                file_contents = toml.load(open_file)
            elif file_type == "json":
                file_contents = json.loads(open_file.read())
            else:
                raise ValueError(f"Unknown file type passed: '{file_type}'")
        logger.debug(f"Successfully loaded local .{file_type} file")
    except Exception as error:
        fail_out(f"Failed to load specified local .{file_type} file: '{local_file_path}'. Exception: {error}")
    return file_contents


def load_contents_from_string(file_type: str, text: str) -> Dict:
    """Loading a local file contents

    Parameters:
        file_type (str) : Type of file to be loaded ie. 'yaml', 'toml', 'json'
        text (str)      : Text string to be loaded as specified filetype
    Returns:
        contents (dict) : The contents of file
    """
    file_type = file_type.lower()
    logger.debug(f"Loading specified text string as filetype '{file_type}' ...")
    if file_type == "yaml":
        contents = yaml.safe_load(text)
    elif file_type == "toml":
        contents = toml.loads(text)
    elif file_type == "json":
        contents = json.loads(text)
    else:
        raise ValueError(f'Unknown file type passed: "{file_type}"')
    logger.debug(f'Successfully loaded specified "{file_type}" contents from text string')
    return contents


def load_contents_from_remote_file_url(file_type: str, remote_file_url: str, allow_redirects: bool = True) -> Dict:
    """Loading a remote yaml file contents over HTTP

    ### FIXME: Make it able to load toml, json, and yaml file types

    Args:
        file_type (str)       : Type of file to be loaded ie. 'yaml', 'toml', 'json'
        remote_file_url (url)  : Remote URL location of file to be loaded
        allow_redirects (bool) : If True allow redirects to another URL (default True)
    Returns:
        file_contents (Dict) : The contents of the file
    """
    file_type = file_type.lower()

    # Getting name of file from URL
    remote_filename = Path(remote_file_url).name
    logger.debug(f"Requested remote filename parsed: {remote_filename}")

    # Check requested file extension
    remote_file_ext = Path(remote_file_url).suffix
    file_ext_accepted = [".yml", ".yaml", ".conf"]
    if not remote_file_ext in file_ext_accepted:
        logger.debug(
            f'Remote file requested "{remote_filename}"" is not one of the accepted file types: {file_ext_accepted}')
        return {}

    # Get request headers
    logger.debug(f'Getting remote file HTTP request headers for "{remote_file_url}" ...')
    try:
        return_content = requests.head(remote_file_url)
    except Exception as error:
        logger.debug(f"Failed to request headers. Exception: {error}")
        return {}
    header = return_content.headers

    # Check if file is below size limit
    content_length = int(header["Content-length"]) / 1000000
    logger.debug(f"Requested file content length: {content_length:.5f} MB)")
    if content_length > 1.0:
        logger.debug(
            f'The requested remote file "{remote_filename}" is {content_length:.2f} MB and larger than 1.0 MB limit, will not download'
        )
        return {}

    # Check if content is text or yaml based
    content_types_accepted = [
        "text/plain",
        "text/x-yaml",
        "application/x-yaml",
        "text/yaml",
        "text/vnd.yaml",
    ]
    content_type = header.get("content-type")
    logger.debug(f"Request content type: {content_type}")
    if not content_type:
        return {}
    elif not any(ext in content_type for ext in content_types_accepted):
        logger.debug(
            f'The content type "{content_type}" of the requested file "{remote_filename}" is not one of the following: {content_types_accepted}'
        )
        return {}

    # Downloading the file content
    logger.debug(f"Requesting remote file: '{remote_file_url}' ...")
    remote_request = requests.get(remote_file_url, allow_redirects=allow_redirects)

    # Check if no error from downloading
    if remote_request.status_code == requests.codes.ok:
        # Loading the yaml file content
        logger.debug("Loading contents of remote file ...")
        try:
            file_contents = yaml.safe_load(remote_request.content)
        except Exception as error:
            logger.debug(f"Failed loading requested file. Exception: {error}")
            return {}
    else:
        logger.debug(
            f"Failed to get remote file '{remote_file_url}'. HTTP request error code {remote_request.status_code}")
        return {}

    return file_contents


def append_lines_to_file(filepath: str, lines_to_append: List[str], location: str = "beginning") -> bool:
    """Add lines to the end to a text based file

    Details: The passed list is parsed and each list item is a separate line added
             to the beginning of the file

    Args:
        filepath        : Path to the file
        lines_to_append : List of strings, each item a line to append
        location        : Location of where to append lines, 'beginning' or 'end' (default 'beginning')

    Returns:
        True if successfully appended, else False
    """
    location = location.lower()

    # Check if file exists
    if not os.path.isfile(filepath):
        logger.debug(f"Failed to find file: {filepath}")
        return False

    logger.debug(f"Appending lines of text to the {location} of file: {filepath} ...")
    try:
        if location == "beginning":
            open_file = open(filepath, "r+")
            lines_old = open_file.readlines()  # read old content
            lines = lines_to_append + lines_old
            open_file.seek(0)
            for line in lines:
                open_file.write(line)
            open_file.close()
        elif location == "end":
            open_file = open(filepath, "a")
            for line in lines_to_append:
                open_file.write(line)
            open_file.close()
        else:
            logger.debug(f"Unsupported append file location: {location}")
            return False
    except Exception as error:
        logger.error(f"Failed to append lines of text to the end of file ({filepath}). Exception: {error}")
        return False
    return True


def is_list_items_in_dict(list_items: list, dict_to_check: dict) -> Union[int, None]:
    """Return index of ANY matched item in the passed list

    Args:
        list_items    : List of items to find in dict_to_check
        dict_to_check : Dict to match the top level keys to

    Example Usage:
        - `is_list_items_in_dict(["hey", "yo"], {"yo": 1}) -> 1`

    Returns:
        Index of any/first matched item in the list to the top level keys of the dict
    """
    for key in dict_to_check:  # Looping top level dict keys
        if key in list_items:  # Check if the key is in one of the list items
            return list_items.index(key)
    return None


def iter_data_empty_item_stripper(iter_data: Union[Dict, List, Set, Tuple]) -> Union[Dict, List, Set, Tuple]:
    """Removes any empty data from a nested or un-nested iter item (list, dict, set, etc)

    Details: https://stackoverflow.com/a/27974027/11294572

    Args:
        iter_data : data in the from of iterable (ie. list, dict, set, etc)

    Returns:
        Iterable item without any blank/empty items
    """
    empties = ((), {}, set(), None)

    if isinstance(iter_data, dict):
        return {
            key: value
            for key, value in ((key, iter_data_empty_item_stripper(value)) for key, value in iter_data.items())
            if value not in empties
        }
    if isinstance(iter_data, list):
        return [value for value in map(iter_data_empty_item_stripper, iter_data) if value not in empties]
    return iter_data


def is_credential_id_format(text: str) -> bool:
    """Checking if the passed text is in Jenkins credential ID format

    Args:
        text: The text to check

    Returns:
        True if the text is in credential ID format, else False
    """
    regex_pattern = (r"^[a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}$")
    cred_match = re.match(regex_pattern, text)
    if cred_match:
        logger.debug(f"Successfully identified credential ID format")
    else:
        logger.debug(f"Failed to identify credential ID format")
    return cred_match


def is_full_url(url: str) -> bool:
    """Check if passed string is a valid and full URL

    Args:
        url: The text to check

    Returns:
        True if valid and full URL, else False
    """
    # TODO: Replace this same function in cli_utility.py usages with this one within classes

    parsed_url = parse_url(url)
    if all([parsed_url.scheme, parsed_url.netloc, parsed_url.path]):
        is_valid_url = True
    else:
        is_valid_url = False
    logger.debug(f"Is valid URL format: {is_valid_url} - {url}")
    logger.debug(f'    - scheme:  {parsed_url.scheme} - {"OK" if parsed_url.scheme else "MISSING"}')
    logger.debug(f'    - netloc:  {parsed_url.netloc} - {"OK" if parsed_url.netloc else "MISSING"}')
    logger.debug(f'    - path:    {parsed_url.path} - {"OK" if parsed_url.path else "MISSING"}')

    return is_valid_url


def url_to_name(url: str) -> str:
    """Convert the jenkins item URL to its name

    Args:
        url : URL of the item (ie. folder, job, build)

    Returns:
        The full name of the item
    """
    # Split url into base components
    url_components = urlparse(url)

    # Split and filter out terms
    # name = url_components.path.strip('/').replace('/job/', '/').strip().strip('/').strip('job/').replace('view/', '').replace('change-requests/', '')
    url_split = url_components.path.strip().strip("/").split("/")
    remove_list = ["job", "view", "change-requests"]
    filtered_list = [list_item for list_item in url_split if list_item not in remove_list]

    # Assemble back
    name = "/".join(filtered_list)

    logger.debug(f'Converted URL "{url}" to fullname "{name}"')
    return name


def format_name(name: str) -> str:
    """Format / clean up the passed name

    Details: The formatting includes it:
        - /Non-PAR/job/Non-Prod-Jobs/job/Something/job/job --> /Non-PAR/Non-Prod-Jobs/Something/job
        - remove `job/`, `view/`, `change-requests/`
        - remove leading or trailing `/`

    Args:
        name : The name of the item

    Returns:
        Formatted item name
    """
    # Filter out terms
    name_formatted = name.strip().replace("/job/", "/").strip("/")
    name_formatted = (name_formatted.replace("job/", "/").strip("/").replace("view/",
                                                                             "").replace("change-requests/", ""))

    logger.debug(f'Formatted "{name}" to "{name_formatted}"')
    return name_formatted


def fullname_to_name(fullname: str) -> str:
    """Convert the jenkins item full name to its name only

    Details: Hey/This/Is/A/Full/Job --> Job

    Args:
        url : Full name of the item (ie. Hey/This/Is/A/Full/Job)

    Returns:
        The name of the item
    """
    name = fullname.strip().strip("/").split("/")[-1]
    logger.debug(f'Converted fullname "{fullname}" to name "{name}"')
    return name


def name_to_url(server_base_url: str, name: str) -> str:
    """Convert the item name to URL

    ** TODO **

    NOTE: Passed '.' will go to base url

    Args:
        name : The name of the item

    Returns:
        Item URL
    """
    # FIXME: Fix this workaround
    cut_off = 1 if name == "." else 0

    a_path = name.strip("/").split("/")
    if len(a_path) > cut_off:
        short_name = "job/" + "/job/".join(a_path)
    else:
        short_name = ""
    # short_name = (('job/' + '/job/'.join(a_path[:-1]) + '/') if len(a_path) > 1 else '')
    url = str(urljoin(server_base_url, short_name))
    logger.debug(f'Converted name "{name}" to URL "{url}"')

    return url


def build_url_to_other_url(build_url: str, target_url: str = "job") -> str:
    """From build_url get job or folder url

    Args:
        build_url  : The URL of the build
        target_url : URL to derive. `job` or `folder` (default: `job`)

    Returns:
        The URL of the job or folder
    """
    # Dissect the build url
    url_parsed = urlparse(build_url)
    path_list = url_parsed.path.split("/")

    # Get the indexes to remove
    if target_url == "job":
        last_index = -2 if not path_list[-1] else -1
    elif target_url == "folder":
        last_index = -4 if not path_list[-1] else -3
    else:
        logger.debug(f"Failed to recognize passed target URL. Passed value: {target_url}")
        return ""

    # Join the new path
    path_new = "/".join(path_list[0:last_index])

    # Assemble the job url
    base_url = url_parsed.scheme + "://" + url_parsed.netloc

    result_url = urljoin(base_url, path_new) + "/"
    logger.debug(f'From build URL "{build_url}" extracted {target_url} URL "{result_url}"')

    return result_url


def build_url_to_build_number(build_url: str) -> Union[int, None]:
    """If possible, extract the build number from a passed Jenkins URL
    In order to extract the build number, the passed URL must be related to the
    build (ie. URL of build logs)

    Args:
        build_url : URL of anything build related

    Example Usage:
        - `build_url_to_build_number("http://......job/myJob/15/") -> 15`
        - `build_url_to_build_number("http://......job/myJob/15/console") -> 15`
        - `build_url_to_build_number("http://......job/myJob") -> None`

    Returns:
        The build number if extracted, else None
    """
    logger.debug(f"Extracting build number from URL: {build_url} ...")
    # Split the URL path, remove the empty items
    url_path_split_list = list(filter(None, urlparse(build_url).path.split("/")))

    # Stepping from the back one item at a time
    build_number = None
    for index in range(len(url_path_split_list) - 1, 2, -1):
        try:
            build_number = int(url_path_split_list[index])
            if url_path_split_list[index - 2] != "job":
                raise ValueError
            break
        except ValueError:
            build_number = None

    logger.debug(f'From build URL "{build_url}" extracted build number "{build_number}"')
    return build_number


def is_complete_build_url(build_url: str) -> bool:
    """Check if passed URL is a complete and valid build URL

    Args:
        build_url : URL of anything build related

    Returns:
        True if valid and complete, else False
    """
    if not build_url:
        return False
    url_path_split_list = list(filter(None, urlparse(build_url).path.split("/")))
    is_complete = True
    try:
        int(url_path_split_list[-1])
        if url_path_split_list[-3] != "job":
            is_complete = False
    except (ValueError, IndexError):
        is_complete = False
    return is_complete


def build_url_complete(build_url: str) -> Union[str, None]:
    """If possible, extract only the build URL from a passed Jenkins URL.
    Will try to truncate anything after the build number.
    In order to extract the complete build URL, the passed URL must be related to the
    build (ie. URL of build logs).

    Args:
        build_url : URL of anything build related

    Example Usage:
        - `build_url_complete("......job/myJob/15/")` -> `......job/myJob/15/``
        - `build_url_complete("......job/myJob/15/console")` -> `......job/myJob/15/`
        - `build_url_complete("......job/myJob/")` -> `None`

    Returns:
        Cleaned build URL, else None
    """
    if not build_url:
        return None

    if is_complete_build_url(build_url):
        return build_url

    logger.debug(f'Extracting build URL from "{build_url}" ...')

    # Split the URL path, remove the empty items
    url_parsed = urlparse(build_url)
    url_path_split_list = list(filter(None, url_parsed.path.split("/")))

    # Stepping from the back, one item at a time
    for index in range(len(url_path_split_list.copy()) - 1, 0, -1):
        try:
            int(url_path_split_list[index])
            if url_path_split_list[index - 2] != "job":
                raise ValueError
            break
        except ValueError:
            url_path_split_list.pop()

    if len(url_path_split_list) < 2:
        logger.debug(f'Cannot extract complete build url from "{build_url}"')
        return None

    # Assemble Everything back
    path_new = "/".join(url_path_split_list)
    base_url = url_parsed.scheme + "://" + url_parsed.netloc
    build_url_complete = urljoin(base_url, path_new) + "/"

    logger.debug(f'Extracted complete build URL "{build_url_complete}" from "{build_url}"')
    return build_url_complete


def item_url_to_server_url(url: str, include_scheme: bool = True) -> str:
    """From build_url get job or folder url

    Args:
        url            : The URL of the build
        include_scheme : If True, include protocol scheme (ie. `https://`) (default: `True`)

    Returns:
        The URL of the server
    """
    url_parsed = urlparse(url)
    if include_scheme:
        server_url = url_parsed.scheme + "://" + url_parsed.netloc
    else:
        server_url = url_parsed.netloc
    return server_url


def has_build_number_started(job_info: dict, build_number: int) -> bool:
    """Given the job info, check if build number has been started/run

    Args:
        job_info     : The job information
        build_number : The build number to be looked up

    Returns:
        True if build number has starated, else false
    """
    if "builds" not in job_info:
        return False
    for build in job_info["builds"]:
        if not "number" in build:
            continue
        if build["number"] == build_number:
            logger.debug(f"Successfully found tha build {build_number} was previously started")
            return True
    logger.debug(f"Failed to find build {build_number} was previously started")
    return False


def item_subitem_list(
    item_info: Dict,
    get_key_info: str,
    item_type: str,
    item_class_list: list = [],
    item_class_key: str = "_class",
) -> Tuple[list, list]:
    """Given a item (job, build, etc) info, get the sub-items matching criteria

    Details: <DETAILED DESCRIPTION>

    Args:
        item_info       : Info dict of the item
        get_key_info    : The key value to get (ie. `url`, `name`)
        item_type       : Class type of sub-item to be matched (ie. `com.cloudbees.hudson.plugins.folder.Folder`) - See `JenkinsItemClasses.py`
        item_class_list : List of classes to be matched (ie. `jobs`, `views`) - See `JenkinsItemClasses.py`
        item_class_key  : The key in the item info to match (ie. `_class`)

    Returns:
        List of the matched item dict and a list of the item names
    """
    # Find subitem info in dict item
    item_list = []
    item_name_list = []
    if item_type in item_info:

        # Loop subsection (ie. jobs, views, etc)
        for subitem_info in item_info[item_type]:

            # Check if subitem has the looked after class
            if subitem_info[item_class_key] in item_class_list:
                item_list.append(subitem_info)
                item_name_list.append(subitem_info[get_key_info])
    else:
        return [], []

    return item_list, item_name_list


def to_seconds(time_quantity: int, time_unit_text: str) -> int:
    """
    Get the number of seconds from the time quantity and time unit type.
    Examples:
        - 45 min -> 2700 seconds
        - 2 days -> 432000 seconds

    Parameters:
        time_quantity (int)  : Number of time units
        time_unit_text (str) : Type of time units (ie. seconds, minutes, hours, etc)
    Returns:
        seconds (int) : Number of seconds
    """
    if not time_quantity:
        return 0

    if time_unit_text in ["s", "sec", "second", "seconds"]:
        return time_quantity

    if time_unit_text in ["m", "min", "minute", "minutes"]:
        return time_quantity * 60

    if time_unit_text in ["h", "hr", "hour", "hours"]:
        return time_quantity * 60 * 60

    if time_unit_text in ["d", "day", "days"]:
        return time_quantity * 60 * 60 * 60

    if time_unit_text in ["blue moon"]:
        blue_moon = 41  # months
        return int(time_quantity * blue_moon * 2.628e6)

    return 0


def html_clean(html: str) -> str:
    """Clean up HTML format to text without HTML tags

    Args:
        html : HTML content

    Returns:
        Cleaned text
    """
    # Remove all HTML tags
    cleaned_text = re.sub(re.compile("<.*?>"), "", html)

    # Convert symbols back
    cleaned_text = cleaned_text.replace("&lt;", "<")
    cleaned_text = cleaned_text.replace("&gt;", ">")
    cleaned_text = cleaned_text.replace("&quot;", '"')
    cleaned_text = cleaned_text.replace("&apos;", "'")
    cleaned_text = cleaned_text.replace("&nbsp;", "")
    cleaned_text = cleaned_text.replace("&amp;", "&")  # This has to be last

    return cleaned_text


def browser_open(url: str, new: int = 2, autoraise: bool = True) -> bool:
    """Clean up HTML format to text without HTML tags

    Args:
        url       : Weblink URL
        new       : 0=Same browser window, 1=New browser window, 2=New Tab
        autoraise : True=Window is raised

    Returns:
        True if successful, else False
    """
    try:
        webbrowser.open(url.strip("/"), new, autoraise)
    except Exception as error:
        logger.debug(f'Failed to open web browser for URL: {url.strip("/")}  Exception: {error}')
        return False
    return True


def has_special_char(text: str, special_chars: str = "@!#$%^&*<>?/\|~:") -> bool:
    """Check if passed text string contains any special characters

    Args:
        text          : Text to check
        special_chars : String with all special characters to check

    Returns:
        True if includes special characters, else False
    """
    regex = re.compile("[" + special_chars + "]")
    includes_special_chars = regex.search(text) != None
    if includes_special_chars:
        logger.debug(f'Item "{text}" includes special characters. Special characters: {special_chars}')
    else:
        logger.debug(f'Item "{text}" does not include special characters. Special characters: {special_chars}')
    return includes_special_chars


def remove_special_char(text: str, special_chars: str = "@!#$%^&*<>?/\|~:") -> str:
    """Remove any special characters from text string

    Args:
        text          : Text to remove special characters from
        special_chars : String with all special characters to remove

    Returns:
        Text with special characters removed
    """
    regex = re.compile("[" + special_chars + "]")
    text_new = re.sub(regex, "", text)
    logger.debug(f'Removed special characters "{special_chars}" from string')
    return text_new


def queue_find(all_queue_info: dict, job_name: str = "", job_url: str = "", first: bool = True) -> list:
    """Finding job in server build queue

    Args:
        TODO

    Returns:
        TODO
    """
    if not job_name and not job_url:
        logger.debug("=No job name or job URL provided")
        return []
    job_name = job_name if job_name else url_to_name(job_url)

    queue_item_matches = []

    for i, queue_item in enumerate(all_queue_info["items"]):
        # Check the item type
        if (queue_item["task"]["_class"] not in JenkinsItemClasses.JOB.value["class_type"]):
            logger.debug(
                f"[ITEM {i+1}/{len(all_queue_info['items'])}] Queued item not a job. Item class: {queue_item['task']['_class']}"
            )
            continue

        queue_job_url = queue_item["task"]["url"]
        logger.debug(f"[ITEM {i+1}/{len(all_queue_info['items'])}] Queue job item: {queue_job_url}")

        queue_job_name = url_to_name(url=queue_job_url)

        # Check for name match
        if queue_job_name == job_name:
            logger.debug(f'Successfully found job in server build queue: "{job_name}"')
            queue_item_matches.append(queue_item)
            if first:
                break

    if not queue_item_matches:
        logger.debug("Failed to find job in build queue")

    return queue_item_matches


def get_resource_path(relative_path: str) -> str:
    """Getting the filepath for existing included resource

    Args:
        relative_path : Relative path within the project directory

    Details:
        - `get_resource_path(os.path.join('resources', 'server_docker_settings', 'last_deploy_info.json'))`
        - `get_resource_path(os.path.join('resources', 'scripts', 'some_script.sh'))`

    Returns:
        Included resource path
    """
    # Get the path in python site packages
    resource_dir = get_project_dir()
    resource_path = os.path.abspath(os.path.join(resource_dir, relative_path))

    # If the file has not been found and it is on windows, try APPDATA directory
    if not os.path.exists(resource_path):
        logger.debug(f'Failed to find resource "{relative_path}" in: {resource_dir}')
        return ""
    logger.debug(f"Successfully found existing resource: {resource_path}")
    return resource_path


def get_project_dir(sample_path: str = "resources") -> str:
    """Getting the path to the directory containing project resources

    Details:
        - Effectively this function is looking through all possible package locations
        and checking if it contains a directory with the project name

    Args:
        sample_path : A directory that is directly inside the project directory (ie. yojenkins/resources)

    Returns:
        Project directory absolute path
    """
    if am_i_bundled():
        # Program is running within a pyinstaller bundle
        project_dir = ""
        possible_dirs = {
            "pyinstaller": sys._MEIPASS,
        }
    else:
        project_dir = "yojenkins"
        possible_dirs = {
            "relative": os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..")),
            "sys_dirs": sysconfig.get_paths()["purelib"],
            "cwd": os.getcwd(),
        }
        # NOTE: "site" module does not work with pyinstaller bundle (AttributeError)
        # 'usr_dirs': site.getusersitepackages(),
        # 'site_dirs': site.getsitepackages(),

    dirs = []
    for possible_dir in possible_dirs.values():
        if isinstance(possible_dir, list):
            dirs.extend(possible_dir)
        else:
            dirs.append(possible_dir)

    logger.debug("Searching project resource directory ...")
    resource_dir_path = ""
    for possible_dir in dirs:
        if os.path.exists(os.path.join(possible_dir, project_dir, sample_path)):
            resource_dir_path = os.path.join(possible_dir, project_dir)
            logger.debug(f"    - {possible_dir} - FOUND")
            break
        logger.debug(f"    - {possible_dir} - NOT FOUND")

    if not resource_dir_path:
        logger.fatal("Failed to find included data directory")
        return ""
    return resource_dir_path


def item_exists_in_folder(item_name: str, folder_url: str, item_type: str, rest: object) -> bool:
    """Checking if the item exists within the specified folder on server

    Args:
        item_name : Name of the item to check
        folder_url : URL of the folder to check
        item_type : Type of the item to check
        rest: Rest object

    Returns:
        True if the item exists, False if not
    """
    item_type_info = getattr(JenkinsItemClasses, item_type.upper())
    prefix = item_type_info.value["prefix"]

    item_url = urljoin(folder_url, f"{prefix}/{item_name}")

    logger.debug(f'Checking if {item_type} "{item_name}" already exists within folder "{folder_url}" ...')
    item_exists = rest.request(f'{item_url.strip("/")}/api/json', "head", is_endpoint=False)[2]
    if item_exists:
        logger.debug(f'Found existing {item_type} "{item_name}" within "{folder_url}"')
        return True
    else:
        logger.debug(f'Did not found {item_type} "{item_name}" within "{folder_url}"')

    return item_exists


def am_i_inside_docker() -> bool:
    """Find out if the program is running inside a docker container

    Returns:
        True if running in docker container, else False
    """
    path = "/proc/self/cgroup"
    return (os.path.exists("/.dockerenv") or os.path.isfile(path) and any("docker" in line for line in open(path)))


def am_i_bundled() -> bool:
    """Find out if the program is running as part of a pyinstaller bundle

    Returns:
        True if running bundled, else False
    """
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def parse_and_check_input_string_list(string_list: str, join_back_char: str = "", split_char: str = ",") -> list:
    """Parsing a string list into a list of strings

    Details:
        - `parse_string_list('a,b,c')` => ['a', 'b', 'c']
        - `parse_string_list('a,b,c', join_back_char=';')` => ['a;b;c']

    Args:
        string_list : String list to be parsed
        split_char : Character to split the string list on
        join_back_char : Character to join the list items

    Returns:
        String with items separated by commas
    """
    parsed_items = []
    for label in string_list.split(split_char):
        label = label.strip()
        if has_special_char(label):
            return []
        parsed_items.append(label)

    if join_back_char:
        parsed_items = join_back_char.join(parsed_items)

    logger.debug(f'Parsed and checked string list: "{string_list}" => "{parsed_items}"')

    return parsed_items


def write_xml_to_file(
    xml_content: str,
    filepath: str,
    opt_json: bool = False,
    opt_yaml: bool = False,
    opt_toml: bool = False,
) -> bool:
    """Writing XML content file in specified format

    Args:
        xml_content : XML content to be written
        filepath    : Filepath to write the content to
        opt_json    : Write the content as JSON
        opt_yaml    : Write the content as YAML
        opt_toml    : Write the content as TOML

    Returns:
        True if successful
    """
    if any([opt_json, opt_yaml, opt_toml]):
        logger.debug("Converting content to JSON ...")
        data_ordered_dict = xmltodict.parse(xml_content)
        content_to_write = json.loads(json.dumps(data_ordered_dict))
    else:
        content_to_write = xml_content  # Keep XML format

    if opt_json:
        content_to_write = json.dumps(data_ordered_dict, indent=4)
    elif opt_yaml:
        logger.debug("Converting content to YAML ...")
        content_to_write = yaml.dump(content_to_write)
    elif opt_toml:
        logger.debug("Converting content to TOML ...")
        content_to_write = toml.dumps(content_to_write)

    logger.debug(f'Writing fetched configuration to "{filepath}" ...')
    try:
        with open(filepath, "w+") as file:
            file.write(content_to_write)
        logger.debug("Successfully wrote configurations to file")
    except Exception as error:
        logger.debug("Failed to write configurations to file. Exception: {error}")
        return False

    return True


def template_apply(string_template: str, is_json: bool = False, **kwargs) -> Union[str, dict]:
    """Apply/Fill variables into a string template.
    Placeholder variables must be in the `${variable_name}` format.

    Details:
        - Example of a string template:
            `'{
                "credentials": {
                    "scope": "${domain}",
                    "username": "${username}"
                }'`

    Args:
        string_template: A string template with placeholders (ie. `${variable}`)
        is_json: If true, the string template is a json string
        kwargs: dictionary of variables to be applied

    Returns:
        String template with variables applied
    """
    logger.debug("Applying variables to string template ...")
    logger.debug(f'Applied variables: {", ".join(list(kwargs.keys()))}')
    # Replace None with empty string
    for key, value in kwargs.items():
        if value is None:
            kwargs[key] = ""

    template = Template(string_template)
    try:
        template_filled = template.safe_substitute(**kwargs)
    except Exception as error:
        logger.debug(f"Failed to apply variables to string template. Exception: {error}")
        return ""
    if is_json:
        try:
            template_filled = json.loads(template_filled)
        except json.JSONDecodeError:
            logger.debug("Failed to parse filled string template as JSON")
            return ""
    logger.debug("Successfully applied variables to string template")
    return template_filled


def run_groovy_script(script_filepath: str, json_return: bool, rest: object,
                      **kwargs) -> Tuple[Union[dict, str], bool, str]:
    """Run a Groovy script on the server and return the response

    Details:
        A failed Groovy script execution will return a list/array in the following format:
        `['yojenkins groovy script failed', '<GROOVY EXCEPTION>', '<CUSTOM ERROR MESSAGE>']`

    Args:
        script_directory: Directory where the script is located
        script_filepath: The path to the Groovy script to run
        json_return: Anticipate and format script return as JSON
        rest: Rest object
        kwargs (dict): Any variables to be inserted into the script text

    Returns:
        Response from the script
        Success flag
        Error message
    """
    logger.debug(f"Loading Groovy script: {script_filepath}")
    try:
        with open(script_filepath, "r") as open_file:
            script = open_file.read()
    except (FileNotFoundError, IOError, PermissionError) as error:
        logger.debug(f"Failed to find or read specified Groovy script file ({script_filepath}). Exception: {error}")
        return (
            {},
            False,
            f"Failed to find or read specified Groovy script file ({script_filepath}). Exception: {error}",
        )

    # Apply passed kwargs to the string template
    if kwargs:
        script = template_apply(string_template=script, is_json=False, **kwargs)
        if not script:
            return {}, False, "Failed to apply variables to Groovy script template"

    # Send the request to the server
    logger.debug(f"Running the following Groovy script on server: {script_filepath} ...")
    script_result, _, success = rest.request(
        target="scriptText",
        request_type="post",
        data={"script": script},
        json_content=False,
    )
    if not success:
        logger.debug("Failed server REST request for Groovy script execution")
        return {}, False, "Failed server REST request for Groovy script execution"

    # Check for yojenkins Groovy script error flag
    if "yojenkins groovy script failed" in script_result:
        groovy_return = eval(script_result.strip(os.linesep))
        logger.debug("Failed to execute Groovy script")
        logger.debug(f"Groovy Exception: {groovy_return[1]}")
        logger.debug(groovy_return[2])
        return (
            {},
            False,
            f"Error while executing Groovy script: {groovy_return[1]}: {groovy_return[2]}",
        )

    # Check for script exception
    exception_keywords = ["Exception", "java:"]
    if any(exception_keyword in script_result for exception_keyword in exception_keywords):
        logger.debug(f"Error keyword matched in script response: {exception_keywords}")
        return (
            {},
            False,
            f"Error keyword matched in script response: {exception_keywords}",
        )

    # Parse script result as JSON
    if json_return:
        try:
            script_result = json.loads(script_result)
        except json.JSONDecodeError as error:
            logger.debug("Failed to parse response to JSON format")
            return {}, False, "Failed to parse response to JSON format"

    return script_result, True, ""


def get_item_action(item_info: dict, class_type: str) -> List[dict]:
    """Get the item's actions for a specific action class type

    Args:
        item_info: The item info dictionary (ie. job info, build info, etc.)
        class_type: The class type of the item

    Returns:
        List of matched actions for the item
    """
    logger.debug(f'Getting actions for item corresponding to class type "{class_type}" ...')
    actions_info = []
    for action in item_info["actions"]:
        if action:
            if action["_class"] == class_type:
                actions_info.append(action)

    return actions_info


def create_new_history_file(file_path: str) -> None:
    """Create a new blank command history file.

    Args:
        file_path: Full path to the history file

    Returns:
        None
    """
    try:
        # Creating configuration directory if it does not exist
        config_dir_abs_path = os.path.join(Path.home(), CONFIG_DIR_NAME)

        if not os.path.exists(config_dir_abs_path):
            logger.debug("Configuration directory does not exist. Creating it ...")
            os.makedirs(config_dir_abs_path)

        if not os.path.exists(file_path):
            logger.debug(f'Command history file NOT found: "{file_path}"')
            logger.debug("Creating a new command history file ...")
        with open(file_path, "w") as open_file:
            json.dump({}, open_file)

    except (FileNotFoundError, IOError, PermissionError) as error:
        fail_out(f"Failed to create history file ({file_path}). Exception: {error}")
    except Exception as error:
        logger.exception(f"Failed to create new command history file. Exception: {error}")


def wait_for_build_and_follow_logs(yj_obj: object, queue_id: int) -> None:
    """Wait for build to leave queue, run, then follow/stream logs to console

    Details:
        Will only use loading spinner when not in logger debug mode

    Args:
        yj_obj:   YoJenkins object
        queue_id: Build queue ID
    """
    msg = f"Build is in queue with queue ID {queue_id}. Waiting for build to run ..."
    if logger.level > 10:
        spinner = yaspin(spinner=getattr(Spinners, "bouncingBar"), attrs=["bold"], text=msg)
        spinner.start()
    else:
        logger.info(msg)

    while True:
        queue_data = yj_obj.job.queue_info(build_queue_number=queue_id)
        if "executable" in queue_data:
            break
        if queue_data.get("stuck"):
            fail_out(
                f"Build is stuck in queue as queue number {queue_id}",
                fg="bright_red",
                bold=True,
            )
        time.sleep(2)

    if logger.level > 10:
        spinner.stop()

    build_number = queue_data["executable"]["number"]
    print2(f"Running with build number {build_number}. Following console logs below:")
    yj_obj.build.logs(
        build_url=None,
        job_url=queue_data["jobUrl"],
        build_number=build_number,
        follow=True,
    )


def diff_show(
    text_1: str,
    text_2: str,
    label_1: str,
    label_2: str,
    line_pattern: tuple,
    char_ignore: int,
    no_color: bool,
    diff_only: bool,
    diff_guide: bool,
) -> None:
    """Display/Show line diffs between two specified texts

    Args:
        text_1:       String text 1
        text_2:       String text 2 to compare to text 1
        label_1:      text_1 label/description
        label_2:      text_2 label/description
        line-pattern: Patterns to consider for diff for each line
        char_ignore:  Number of characters to ignore for comparison at start of each line
        no_color:     Display with no color
        diff_only:    Only show lines that are different
        diff_guide:   Show diff specifiers/guides to show where difference is on line
    """
    text_1, text_2 = text_1.splitlines(), text_2.splitlines()

    # Ignore specified number of initial characters
    if char_ignore > 0:
        logger.debug(f"Applying {char_ignore} initial characters for each line before diff ...")
        text_1 = [line[char_ignore:] for line in text_1]
        text_2 = [line[char_ignore:] for line in text_2]

    # Only select REGEX line patterns to diff
    if line_pattern:
        regex_pattern = "|".join(list(line_pattern))
        logger.debug(f'Applying REGEX pattern line filter before diff "{regex_pattern}":')
        count_1, count_2 = len(text_1), len(text_2)

        text_1 = [re.findall(regex_pattern, line) for line in text_1]  # Apply regex
        text_1 = ["".join(line) for line in filter(lambda x: x, text_1)]  # Filter
        logger.debug(f"   - Text 1: Kept {len(text_1)} of {count_1} lines ({len(text_1)/count_1 * 100:.1f}%)")

        text_2 = [re.findall(regex_pattern, line) for line in text_2]
        text_2 = ["".join(line) for line in filter(lambda x: x, text_2)]
        logger.debug(f"   - Text 2: Kept {len(text_2)} of {count_2} lines ({len(text_2)/count_2 * 100:.1f}%)")

    # Compute the diff
    # lines_diff = difflib.Differ().compare(text_1, text_2)
    # lines_diff = difflib.unified_diff(text_1, text_2, fromfile="Build URL 1", tofile="Build URL 2")
    lines_diff = difflib.ndiff(text_1, text_2)
    diff_ratio = difflib.SequenceMatcher(None, text_1, text_2).quick_ratio() * 100

    logger.debug("Showing the diff of two provided text strings ...")
    logger.debug("Diff output options specified:")
    logger.debug(f"    - Show no color:   {no_color}")
    logger.debug(f"    - Show diff only:  {diff_only}")
    logger.debug(f"    - Show diff guide: {diff_guide}")

    # Legend Header
    print()
    if no_color:
        print(label_1)
        print(label_2)
    else:
        secho(style(label_1, fg="red", bold=False))
        secho(style(label_2, fg="green", bold=False))

    if char_ignore > 0:
        print(f"***  NOTE: Ignoring first {char_ignore} characters of each line")
    if line_pattern:
        print(f"***  NOTE: Only considering log lines with REGEX pattern: {regex_pattern}")
    print("")
    print("-" * 51)

    for line in lines_diff:
        if line.isspace():  # difflib sometimes returns empty strings
            continue
        if diff_only and line[0] not in ["-", "+", "?"]:  # Skip non-diff lines
            continue
        if not diff_guide and line.startswith("?"):
            continue

        if no_color:
            color, bold = "reset", False
        elif line[0] == "+":
            color, bold = "green", False
        elif line[0] == "-":
            color, bold = "red", False
        elif line[0] == "?":
            color, bold = "yellow", True
        else:
            color, bold = None, False

        if line.startswith("?"):
            line = line.replace("?", " ", 1)

        secho(style(f"  {line}", fg=color, bold=bold))

    print("-" * 51)
    print(f"\n  Similarity: {diff_ratio:.1f}%")



"""Build class definition"""

import difflib
import logging
import os
from datetime import datetime, timedelta
from itertools import islice
from time import sleep, time
from typing import Dict, List, Tuple
from urllib.parse import urlencode

import click
import requests
import yaml

from yojenkins.monitor import BuildMonitor
from yojenkins.utility import utility
from yojenkins.utility.utility import diff_show, fail_out, failures_out, print2
from yojenkins.yo_jenkins.auth import Auth
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses
from yojenkins.yo_jenkins.rest import Rest
from yojenkins.yo_jenkins.status import BuildStatus

# Getting the logger reference
logger = logging.getLogger()


class Build():
    """Buld class"""

    def __init__(self, rest: Rest, auth: Auth) -> None:
        """Object constructor method, called at object creation

        Args:
            rest: Rest object
            auth: Auth object

        Returns:
            None
        """
        self.rest = rest
        self.auth = auth
        self.build_monitor = BuildMonitor(rest, auth, self)

        self.build_logs_extension = ".log"

    def info(self,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: int = None,
             latest: bool = False) -> Dict:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            build_url = utility.build_url_complete(build_url)
            request_url = f"{build_url.strip('/')}/api/json"
            build_info = self.rest.request(request_url, 'get', is_endpoint=False)[0]
            if not build_info:
                fail_out(f'Failed to get build info for provided build url ({build_url})')
        else:
            if not job_name and not job_url:
                fail_out('No job name, job url, and build url provided')

            if job_name and not job_url:
                job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

            job_info, _, success = self.rest.request(f'{job_url.strip("/")}/api/json', 'get', is_endpoint=False)
            if not success:
                fail_out(f'Failed getting build info, because failed to request job info: {job_url}')

            # Check if found item type/class is a build
            if job_info['_class'] not in JenkinsItemClasses.JOB.value['class_type']:
                fail_out(f'Failed to match job type/class. The found item is "{job_info["_class"]}"')

            try:
                job_last_build_number = job_info['lastBuild']['number'] if 'lastBuild' in job_info else 0
            except TypeError:
                fail_out('Failed to find previous builds. This job may not have any past builds')

            # If build number is not passed, get the latest build number for job
            if not build_number and latest:
                # Build number not passed and latest flag is not set
                logger.debug(
                    f'No build number passed BUT --latest flag set. Using latest build number for this job: {job_last_build_number}'
                )
                build_number = job_last_build_number
            elif not build_number and not latest:
                fail_out('Failed to specify build. No build number passed and --latest flag not set')
            else:
                # Build number is passed
                if build_number > job_last_build_number:
                    fail_out('Failed to specify build. Build number exceeds last build number for this job')

            logger.debug(f'Getting build info for job "{job_info["fullName"]}, build {build_number} ...')
            build_info, _, success = self.rest.request(f'{job_url.strip("/")}/{build_number}/api/json',
                                                       'get',
                                                       is_endpoint=False)
            if not success:
                fail_out('Failed to request build info')

        # Check if found item type/class is a build
        if build_info['_class'] not in JenkinsItemClasses.BUILD.value['class_type']:
            fail_out(f'Build found, but failed to match build type/class. This item is "{build_info["_class"]}"')

        # Add additional derived information
        if 'timestamp' in build_info:
            build_info['startDatetime'] = datetime.fromtimestamp(build_info['timestamp'] /
                                                                 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
            build_info['estimatedDurationFormatted'] = str(timedelta(
                seconds=build_info["estimatedDuration"] /
                1000.0))[:-3] if build_info["estimatedDuration"] > 0 else None

            # Check if results are in
            if 'result' in build_info:
                if build_info['result']:
                    build_info['resultText'] = build_info['result']
                    build_info['durationFormatted'] = str(timedelta(seconds=build_info['duration'] / 1000.0))[:-3]
                    build_info['endDatetime'] = datetime.fromtimestamp(
                        (build_info['timestamp'] + build_info['duration']) / 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
                    build_info['elapsedFormatted'] = build_info['durationFormatted']
                else:
                    build_info['resultText'] = BuildStatus.RUNNING.value
                    build_info['durationFormatted'] = None
                    build_info['endDatetime'] = None
                    build_info['elapsedFormatted'] = str(timedelta(seconds=(time() -
                                                                            build_info['timestamp'] / 1000)))[:-3]

            else:
                build_info['resultText'] = BuildStatus.UNKNOWN.value
        else:
            build_info['startDatetime'] = None
            build_info['estimatedDurationFormatted'] = None
            build_info['resultText'] = BuildStatus.NOT_RUN.value
            build_info['durationFormatted'] = None
            build_info['endDatetime'] = None
            build_info['elapsedFormatted'] = None

        if 'url' in build_info:
            build_info['fullName'] = utility.url_to_name(build_info['url'])
            build_info['jobUrl'] = utility.build_url_to_other_url(build_info['url'], target_url='job')
            build_info['jobFullName'] = utility.url_to_name(build_info['jobUrl'])
            build_info['jobName'] = utility.fullname_to_name(build_info['jobFullName'])
            build_info['folderUrl'] = utility.build_url_to_other_url(build_info['url'], target_url='folder')
            build_info['folderFullName'] = utility.url_to_name(build_info['folderUrl'])
            build_info['folderName'] = utility.fullname_to_name(build_info['folderFullName'])
            build_info['serverURL'] = utility.item_url_to_server_url(build_info['url'])
            build_info['serverDomain'] = utility.item_url_to_server_url(build_info['url'], False)

            build_info[
                'folderFullName'] = 'Base Folder' if not build_info['folderFullName'] else build_info['folderFullName']

        if 'builtOn' not in build_info:
            build_info['builtOn'] = 'N/A'
        else:
            build_info['builtOn'] = 'N/A' if not build_info['builtOn'] else build_info['builtOn']

        return build_info

    def status_text(self,
                    build_url: str = '',
                    job_name: str = '',
                    job_url: str = '',
                    build_number: int = None,
                    latest: bool = False) -> str:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Get the build info
        build_url = utility.build_url_complete(build_url)
        build_info = self.info(build_url=build_url,
                               job_name=job_name,
                               job_url=job_url,
                               build_number=build_number,
                               latest=latest)

        # If nothing is returned, check if job is queued on server
        logger.debug('The specified build was not found in job')
        logger.debug('Looking for build in the server build queue ...')
        if build_url:
            job_url = utility.build_url_to_other_url(build_url)
        elif job_name:
            pass
        elif job_url:
            pass
        else:
            fail_out('Failed to find build status text. Specify build url, job name, or job url')
        logger.debug(f'Job name: {job_name}')

        # Requesting all queue and searching queue (NOTE: Could use Server object)
        logger.debug(f'Requesting all build queue items ...')
        queue_all = self.rest.request('queue/api/json', 'get')[0]
        logger.debug(f"Number of queued items found: {len(queue_all['items'])}")
        queue_matches = utility.queue_find(queue_all, job_name=job_name, job_url=job_url)
        if not queue_matches:
            fail_out('Failed to find running or queued builds')
        queue_info = queue_matches[0]

        if not queue_info:
            logger.debug('Build for job NOT found in queue')
            return BuildStatus.NOT_FOUND.value
        else:
            logger.debug(f'Build for job found in queue. Queue number {queue_info["id"]}')
            return BuildStatus.QUEUED.value

        # FIXME: resultText is returned in build info. Maybe move queue check to build_info??

        # Check if in process (build is there but results not posted)
        if 'result' not in build_info:
            logger.debug('Build was found running/building, however no results are posted')
            return BuildStatus.RUNNING.value
        else:
            # FIXME: Get "No status found" when "yojenkins build status --url" on build that is "RUNNING" (result: Null)
            logger.debug('Build found, but has concluded or stopped with result')
            return build_info['result']

    def abort(self,
              build_url: str = '',
              job_name: str = '',
              job_url: str = '',
              build_number: int = None,
              latest: bool = False) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            build_url = utility.build_url_complete(build_url)
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Aborting build: {url} ...')
        request_url = f"{url.strip('/')}/stop"
        if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
            messages = [
                'Failed to abort build', '   - Build may not exist, is queued, or you are not authorized to abort',
                '   - For more details run with --debug'
            ]
            failures_out(messages)

        logger.debug('Successfully aborted build')

        return utility.build_url_to_build_number(build_url=url)

    def delete(self,
               build_url: str = '',
               job_name: str = '',
               job_url: str = '',
               build_number: int = None,
               latest: bool = False) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Pass a list of build numbers
        if build_url:
            build_url = utility.build_url_complete(build_url)
            logger.debug(f'Build URL passed: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Deleting build: {url} ...')
        request_url = f"{url.strip('/')}/doDelete"
        if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
            fail_out('Failed to delete build. Build may not exist or is queued')

        return utility.build_url_to_build_number(build_url=url)

    def stage_list(self,
                   build_url: str = '',
                   job_name: str = '',
                   job_url: str = '',
                   build_number: int = None,
                   latest: bool = False) -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # FIXME: yojenkins build stages --url https://localhost:8080/job/Non-PAR/job/Non-Prod-Jobs/job/Something/job/test_job/46/
        #        yields 404 in running build. Maybe issue with formatting of the url to name?

        # TODO: Pass a list of build numbers
        if not build_url:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_url = utility.build_url_complete(build_url)
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            build_url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Getting build stages for: {build_url} ...')
        request_url = f"{build_url.strip('/')}/wfapi/describe"
        return_content, _, return_success = self.rest.request(request_url, 'get', is_endpoint=False)
        if not return_success or not return_content:
            fail_out('Failed to get build stages. This may not be a staged build')

        # Getting the stage items
        # FIXME: When --url <job> and no build number is passed, it will just get the job describe, not build info
        if 'stages' in return_content:
            build_stage_list = return_content['stages']
        else:
            fail_out('Failed to find "stages" key in build info. This may not be a staged build')

        # Add additional derived information for each step
        for stage_info in build_stage_list:
            stage_info['startDatetime'] = datetime.fromtimestamp(stage_info["startTimeMillis"] /
                                                                 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
            stage_info['durationFormatted'] = str(timedelta(seconds=stage_info["durationMillis"] / 1000.0))[:-3]
            stage_info['pauseDurationFormatted'] = str(timedelta(seconds=stage_info["pauseDurationMillis"] / 1000.0))
            stage_info['url'] = stage_info['_links']['self']['href']

        # Getting only the names of the stages
        build_stage_name_list = [stage['name'] for stage in build_stage_list]

        return build_stage_list, build_stage_name_list

    def artifact_list(self,
                      build_url: str = '',
                      job_name: str = '',
                      job_url: str = '',
                      build_number: int = None) -> List[dict]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Test on build with artifacts
        build_url = utility.build_url_complete(build_url)
        return self.info(build_url=build_url, job_name=job_name, job_url=job_url,
                         build_number=build_number).get('artifacts')

    def artifact_download(self):
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Test on build with artifacts
        pass

    def logs(self,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: int = None,
             latest: bool = False,
             tail: float = None,
             download_dir: str = '',
             follow: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            build_url = utility.build_url_complete(build_url)
        else:
            logger.debug('No build URL passed. Getting build information through job ...')
            build_info = self.info(job_name=job_name, job_url=job_url, build_number=build_number, latest=latest)
            build_url = build_info['url']

        # FIXME: Check if this is an actual build and job/folder/etc

        request_url = f"{build_url.strip('/')}/consoleText"

        if download_dir:
            # Download to local file
            auth = requests.auth.HTTPBasicAuth(self.rest.username, self.rest.api_token)
            filename = f'build-logs_{datetime.now().strftime("%m-%d-%Y_%I-%M-%S")}{self.build_logs_extension}'
            logger.debug(f'Downloading console text logs to local file "{filename}" ...')
            try:
                with requests.get(request_url, auth=auth, stream=True, timeout=10) as open_request:
                    open_request.raise_for_status()
                    with open(os.path.join(download_dir, filename), 'wb') as open_file:
                        for chunk in open_request.iter_content(chunk_size=8192):
                            if chunk:
                                open_file.write(chunk)
                logger.debug('Successfully download build logs to file')
            except Exception as error:
                fail_out(f'Failed to download or save logs for build. Exception: {error}')
        else:
            if not follow:
                # Show build logs in console
                logger.debug('Fetching logs from server ...')
                return_content, _, return_success = self.rest.request(request_url,
                                                                      'get',
                                                                      is_endpoint=False,
                                                                      json_content=False)
                if not return_success or not return_content:
                    fail_out('Failed to get console logs. Build may not exist or is queued')

                # If tail/last part of the log was specified
                if tail:
                    logger.debug(f'--tail option specified with value of: {tail}')
                    tail = abs(tail)
                    logs_list = list(map(lambda num: num.strip(), return_content.splitlines()))
                    number_of_lines = round(len(logs_list) * tail) if tail < 1 else round(tail)
                    start_log_number = 0 if number_of_lines > len(logs_list) else len(logs_list) - number_of_lines
                    return_content = os.linesep.join(list(islice(logs_list, start_log_number, None)))
                    logger.debug(f'Only printing out last logs, lines {start_log_number} to {len(logs_list)} ...')

                logger.debug('Printing out console text logs ...')
                print2(return_content)
            else:
                # Stream the logs to console
                log_poll_interval = 1.0
                logger.debug(f'Following/streaming logs from server at poll interval: {log_poll_interval}s ...')

                # Check if Jenkins server supports progressiveText
                _, headers, request_success = self.rest.request(
                    f"{build_url.strip('/')}/logText/progressiveText?start=0",
                    'head',
                    is_endpoint=False,
                    json_content=False)

                if request_success and "X-Text-Size" in headers:
                    # METHOD 1: Fetch logs using progressiveText endpoint
                    logger.debug('Jenkins server supports requesting partial byte ranges (progressiveText)')
                    try:
                        progressive_text_position = headers['X-Text-Size']
                        while True:
                            request_url = f"{build_url.strip('/')}/logText/progressiveText?start={progressive_text_position}"
                            return_content, headers, _ = self.rest.request(request_url,
                                                                           'get',
                                                                           is_endpoint=False,
                                                                           json_content=False)
                            logger.debug(f'Request Headers: {headers}')
                            if len(return_content) != 0:
                                print2(return_content.strip())
                            if 'X-More-Data' not in headers:
                                break
                            progressive_text_position = headers['X-Text-Size']
                            sleep(log_poll_interval)
                    except KeyboardInterrupt:
                        logger.debug('Keyboard Interrupt (CTRL-C) by user. Stopping log following ...')
                else:
                    # METHOD 2: Fetch logs using log difference (Server does not support progressiveText)
                    try:
                        logger.debug(
                            'Jenkins server does not support requesting partial byte ranges (progressiveText), '
                            'MUST download entire log to get log message differences')
                        old_dict = {}
                        fetch_number = 1
                        request_url = f"{build_url.strip('/')}/consoleText"
                        while True:
                            headers = self.rest.request(request_url, 'head', is_endpoint=False, json_content=False)[1]
                            if 'content-length' not in headers:
                                fail_out(f'Failed to find "content-length" key in server response headers: {headers}')
                            content_length_sample_1 = int(headers['Content-Length'])
                            sleep(1)
                            headers = self.rest.request(request_url, 'head', is_endpoint=False, json_content=False)[1]
                            content_length_sample_2 = int(headers['Content-Length'])

                            content_length_diff = content_length_sample_2 - content_length_sample_1
                            if content_length_diff:
                                logger.debug(
                                    f'LOG FETCH {fetch_number}: Content length difference: {content_length_diff} bytes'
                                )
                                return_content = self.rest.request(request_url,
                                                                   'get',
                                                                   is_endpoint=False,
                                                                   json_content=False)[0]
                                new_dict = dict.fromkeys(
                                    list(map(lambda num: num.strip(), return_content.splitlines())))

                                diff = dict.fromkeys(x for x in new_dict if x not in old_dict)
                                diff_list = list(diff.keys())
                                diff_text = os.linesep.join(diff_list)

                                old_dict = new_dict
                                fetch_number += 1

                                print2(diff_text)
                            else:
                                logger.debug('No content length difference')
                    except KeyboardInterrupt:
                        logger.debug('Keyboard Interrupt (CTRL-C) by user. Stopping log following ...')
        return True

    def browser_open(self,
                     build_url: str = '',
                     job_name: str = '',
                     job_url: str = '',
                     build_number: int = None,
                     latest: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Need URL
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            build_url = utility.build_url_complete(build_url)
            build_url = build_url.strip('/')
        else:
            logger.debug('No build URL passed. Getting build information ...')
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            build_url = build_info['url']

        # Open the build in browser
        logger.debug(f'Opening build in web browser: "{build_url}" ...')
        success = utility.browser_open(url=build_url)
        if not success:
            fail_out('Failed to open build in web browser')
        logger.debug('Successfully opened build in web browser')

        return success

    def monitor(self,
                build_url: str = '',
                job_name: str = '',
                job_url: str = '',
                build_number: int = None,
                latest: bool = False,
                sound: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            build_url = utility.build_url_complete(build_url)
            if not self.rest.request(f"{build_url.strip('/')}/api/json", 'head', is_endpoint=False)[2]:
                fail_out(f'Failed to find build. The build may not exist: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        logger.debug(f'Starting monitor for: "{url}" ...')
        success = self.build_monitor.monitor_start(build_url=url, sound=sound)
        if not success:
            fail_out('Failed to start build monitor')
        logger.debug('Successfully started build monitor')

        return success

    def parameters(self,
                   build_url: str = '',
                   job_name: str = '',
                   job_url: str = '',
                   build_number: int = None,
                   latest: bool = False) -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Pass a list of build numbers
        build_url = utility.build_url_complete(build_url)
        build_info = self.info(build_url, job_name, job_url, build_number, latest)
        if not build_url:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_url = build_info['url']

        logger.debug(f'Getting build parameters for: "{build_url}" ...')

        # Check if item has any parameter actions
        parameter_actions = utility.get_item_action(build_info, 'hudson.model.ParametersAction')
        if not parameter_actions:
            fail_out('This build does not have any build parameters')

        # Get the parameter definitions
        parameters = parameter_actions[0]['parameters']

        # List of parameter items
        parameters_list = [
            f'{parameter["_class"].split(".")[-1]} - {parameter["name"]} - {parameter["value"]}'
            for parameter in parameters
        ]

        return parameters, parameters_list

    def rebuild(self,
                build_url: str = '',
                job_name: str = '',
                job_url: str = '',
                build_number: int = None,
                latest: bool = False) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        build_url = utility.build_url_complete(build_url)
        build_info = self.info(build_url, job_name, job_url, build_number, latest)
        if not build_url:
            build_url = build_info['url']
        if not job_url:
            job_url = utility.build_url_to_other_url(build_url, "job")

        logger.debug(f'Rebuilding build: "{build_info["url"]}" ...')
        parameter_actions = utility.get_item_action(build_info, 'hudson.model.ParametersAction')
        try:
            parameters = parameter_actions[0]['parameters']
            parameters = {param['name']: param['value'] for param in parameters}
            logger.debug(f'Triggering job with job build parameters: {parameters}')
            post_url = f'{job_url}/buildWithParameters?{urlencode(parameters)}'
        except IndexError:
            logger.debug('No job build parameters found for this build')
            post_url = f'{job_url}/build'

        return_headers = self.rest.request(post_url, 'post', is_endpoint=False)[1]

        # Parse the queue location of the build
        if return_headers:
            build_queue_url = return_headers['Location']
            if build_queue_url.endswith('/'):
                queue_location = build_queue_url[:-1]
            parts = queue_location.split('/')
            build_queue_number = int(parts[-1])
            logger.debug(f'Build queue URL: {queue_location}')
            logger.debug(f'Build queue ID: {build_queue_number}')
        else:
            fail_out('Failed to trigger build. Failed to send request')

        return build_queue_number

    def diff(
        self,
        build_url_1: str = '',
        build_url_2: str = '',
        logs: bool = False,
        line_pattern: tuple = (),
        char_ignore: int = 0,
        no_color: bool = False,
        diff_only: bool = False,
        diff_guide: bool = False,
    ) -> None:
        """Get the diff comparison for two builds

        Args:
            build_url_1:  First build for comparison
            build_url_2:  Second build for comparison
            logs:         Compare build logs
            line-pattern: Patterns to consider for diff for each line
            char_ignore:  Number of characters to ignore at start of each line
            no_color:     Output diff with no color
            diff_only:    Only show the lines that have changed
            diff_guide:   Show diff guide, showing where exactly difference is in line
        """
        build_url_1 = utility.build_url_complete(build_url_1)
        if not build_url_1:
            fail_out('Failed to parse provided BUILD_URL_1. Please check specified arguments')
        build_url_2 = utility.build_url_complete(build_url_2)
        if not build_url_2:
            fail_out('Failed to parse provided BUILD_URL_2. Please check specified arguments')

        logger.debug(f'Getting build {"LOGS" if logs else "INFO"} diff for the following two builds:')
        logger.debug(f'    - Build 1:   {build_url_1}')
        logger.debug(f'    - Build 2:   {build_url_2}')

        if logs:
            build_logs_1, _, success = self.rest.request(f"{build_url_1.strip('/')}/consoleText",
                                                         'get',
                                                         is_endpoint=False,
                                                         json_content=False)
            if not success:
                fail_out(f'Failed to fetch logs for build "{build_url_1}"')

            build_logs_2, _, success = self.rest.request(f"{build_url_2.strip('/')}/consoleText",
                                                         'get',
                                                         is_endpoint=False,
                                                         json_content=False)
            if not success:
                fail_out(f'Failed to fetch logs for build "{build_url_2}"')

            diff_show(build_logs_1, build_logs_2, "---  BUILD 1", "+++  BUILD 2", line_pattern, char_ignore, no_color,
                      diff_only, diff_guide)
        else:
            build_info_1 = self.info(build_url=build_url_1)
            build_info_2 = self.info(build_url=build_url_2)
            build_info_yaml_1 = yaml.safe_dump(build_info_1, default_flow_style=False, indent=2)
            build_info_yaml_2 = yaml.safe_dump(build_info_2, default_flow_style=False, indent=2)
            diff_show(build_info_yaml_1, build_info_yaml_2, "---  BUILD 1", "+++  BUILD 2", line_pattern, char_ignore,
                      no_color, diff_only, diff_guide)
